# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['DOT', 'IPYTHON_ATTRIBUTES', 'PROTOCOL_META_ATTRS', 'ANNOYING_ATTRS', 'MODULE_TYPE_ATTRS', 'DUCK_SPEC_ATTRS']

# %% ../nbs/00_cons.ipynb 6
#| export


# %% ../nbs/00_cons.ipynb 8
#| export


# %% ../nbs/00_cons.ipynb 10
#| export

# %% ../nbs/00_cons.ipynb 12
#| export

# %% ../nbs/00_cons.ipynb 14
#| export

# %% ../nbs/00_cons.ipynb 16
DOT = '.'

# %% ../nbs/00_cons.ipynb 17
__HASH__ = '__hash__'

# %% ../nbs/00_cons.ipynb 18
IPYTHON_ATTRIBUTES = {
    '_ipython_canary_method_should_not_exist_', 
    '_ipython_display_',
    '_repr_mimebundle_',
    '_repr_html_',
    '_repr_markdown_',
    '_repr_svg_',
    '_repr_png_',
    '_repr_pdf_',
    '_repr_jpeg_',
    '_repr_latex_',
    '_repr_json_',
    '_repr_javascript_'
}
'''Attributes that `ipython` might invoke on an object when used in a notebook or console environment.''';

PROTOCOL_META_ATTRS = {'_is_protocol', '_is_runtime_protocol'}
'''Attributes that are invoked when `isinstance` is called on a ProtocolMeta subclass.''';

ANNOYING_ATTRS = IPYTHON_ATTRIBUTES | PROTOCOL_META_ATTRS
'''Attributes that are annoying to deal with when introspecting an object and should be ignored.''';

MODULE_TYPE_ATTRS = set({'__name__', '__doc__', '__loader__', '__spec__'})
'''Attributes that are required to represent a ModuleType''';

# %% ../nbs/00_cons.ipynb 19
DUCK_SPEC_ATTRS = ('asset', 'is_module', 'ducked', 'default', 'bases', 'attrs', 'delayed')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_base.ipynb.

# %% auto 0
__all__ = ['Duck', 'DuckModule', 'DuckSpec']

# %% ../nbs/09_base.ipynb 6
import sys

# %% ../nbs/09_base.ipynb 8
from types import ModuleType, FunctionType, BuiltinFunctionType
from typing import Any, Self, Union, Callable

# %% ../nbs/09_base.ipynb 10
#| export

# %% ../nbs/09_base.ipynb 12
#| export

# %% ../nbs/09_base.ipynb 14
from .cons import DUCK_SPEC_ATTRS
from .type import P
from .util import (prep_kwargs, asset_last, tryattr)
from .meta import DuckModuleMeta, ABCDuck

# %% ../nbs/09_base.ipynb 17
class Duck(metaclass=ABCDuck):
    '''A class representing a duck-type.'''

# %% ../nbs/09_base.ipynb 22
class DuckModule(ModuleType, metaclass=DuckModuleMeta):
    ...

# %% ../nbs/09_base.ipynb 30
class DuckSpec(object):
    asset: str
    is_module: bool = False
    ducked: Union['DuckModule',  None] = None
    default: Any = None
    bases: tuple | Callable[[], tuple] = ()
    attrs: dict | Callable[[], dict] = {}
    delayed: dict = {}
    
    def __prep__(self, *args: P.args, **kwargs: P.kwargs) -> dict:
        # prepare arguments by overwriting classvars with *args / **kwargs
        return prep_kwargs(self, *args, __attrs=DUCK_SPEC_ATTRS, **kwargs)
    
    def __load__(self, *args: P.args, **kwargs: P.kwargs) -> type:
        # prepare arguments by overwriting classvars with *args / **kwargs
        kwds = self.__prep__(*args, **kwargs)
        
        # try to load the asset as a module or class
        item = tryattr(**kwds)
        
        # try to preserve the duckspec by storing it in the item
        try: setattr(item, 'duckspec', self)
        except: ...
        
        # update the duckspec with the item
        setattr(self, '_duck', item)
        return item
    
    def __new__(cls, *args: P.args, **kwargs: P.kwargs) -> Self:
        inst = super().__new__(cls)
        kwds = inst.__prep__(*args, **kwargs)
        name = asset_last(kwds.get('asset', ''))
        # save instance variables
        for k, v in kwds.items(): setattr(inst, k, v)
        return inst
    
    def __init__(
        self: Self, 
        asset: str, 
        is_module: bool = False, 
        ducked: Union['DuckModule', None] = None,
        default: Any = None, 
        bases: tuple | Callable[[], tuple] = (),
        attrs: dict | Callable[[], dict] = {},
        **kwargs
    ):
        kwds = dict(asset=asset, is_module=is_module, ducked=ducked, default=default, bases=bases, attrs=attrs)
        for k, v in kwds.items(): setattr(self, k, v)
            
    def __repr__(self: Self) -> str:
        return f'''<duckspec {getattr(self, 'asset', '')}>'''
    
    @property
    def duck(self: Self) -> Union[type, ModuleType, FunctionType, BuiltinFunctionType]:
        try: 
            return self._duck
        except AttributeError: 
            try:
                self._duck = self.__load__()
            except Exception as e:
                print(f'Error loading duckspec {self}: {e}')
                return self.default
        
        return self._duck
    
    def load(self: Self) -> Union[type, ModuleType, FunctionType, BuiltinFunctionType]: 
        if hasattr(self, '_duck'): return self._duck
        return self.duck 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/11_mods.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/11_mods.ipynb 6
#| export


# %% ../nbs/11_mods.ipynb 8
from types import ModuleType

# %% ../nbs/11_mods.ipynb 10
#| export

# %% ../nbs/11_mods.ipynb 12
#| export

# %% ../nbs/11_mods.ipynb 14
from .base import DuckModule

# %% ../nbs/11_mods.ipynb 16
class _os(DuckModule): 
    __duckmodule__ = 'os'
    __alts__ = dict(PathLike=str)

class _pathlib(DuckModule): 
    __duckmodule__ = 'pathlib'
    __alts__ = dict(Path=str, PosixPath=str)
    
class _types(DuckModule):
    __duckmodule__ = 'types'
    __alts__ = dict(ModuleType=type, MethodWrapperType=type, BuiltinFunctionType=type, TracebackType=type, FunctionType=type, NoneType=type(None))
    
class _re(DuckModule):
    __duckmodule__ = 're'
    __alts__ = dict(Pattern=type, Match=type)
    
class _string(DuckModule):
    __duckmodule__ = 'string'
    __alts__ = dict(Template=type, Match=type)
    
class _math(DuckModule): 
    __duckmodule__ = 'math'
    __alts__ = dict(nan=float('nan'))
    
class _numbers(DuckModule): 
    __duckmodule__ = 'numbers'
    __alts__ = dict(Number=type)
    
class _inspect(DuckModule):
    __duckmodule__ = 'inspect'
    __alts__ = dict(Signature=type, Parameter=type, ParameterKind=type, BoundArguments=type)
    
class _importlib(DuckModule):
    __duckmodule__ = 'importlib'
    __alts__ = dict(Loader=type, ModuleSpec=type)
    
class _itertools(DuckModule):
    __duckmodule__ = 'itertools'
    __alts__ = dict(chain=type, product=type, combinations=type, permutations=type, zip_longest=type)
    
class _logging(DuckModule):
    __duckmodule__ = 'logging'
    __alts__ = dict(Logger=type, LogRecord=type, Handler=type, Filter=type, Formatter=type)
    
class _collectionsabc(DuckModule):
    __duckmodule__ = '_collections_abc'
    __alts__ = dict(dict_keys=type, dict_values=type, dict_items=type)
    
class _numpy(DuckModule): 
    __duckmodule__ = 'numpy'
    __alts__ = dict(
        nan=float('nan'), NaN=float('nan'), NAN=float('nan'),
        bool_=bool, dtype=type,
        ndarray=list, matrix=list, recarray=list,
        DataSource=str, empty=None
    )
    
class _scipy(DuckModule):
    __duckmodule__ = 'scipy'
    __alts__ = dict(
        sparray=list, spmatrix=list, 
        cscarray=list, cscmatrix=list,
        cooarray=list, coomatrix=list,
        csrarray=list, csrmatrix=list,
        diaarray=list, diamatrix=list,
        dokarray=list, dokmatrix=list,
        lilarray=list, lilmatrix=list,
        bsrarray=list, bsrmatrix=list,
    )
    
class _pandas(DuckModule):
    __duckmodule__ = 'pandas'
    __alts__ = dict(
        NA=float('nan'), Index=list, DataFrame=list, Series=list, CategoricalIndex=list,
        PandasDelegate=object, CategoricalAccessor=object
    )
    
class _sklearn(DuckModule):
    __duckmodule__ = 'sklearn'
    __alts__ = dict(BaseEstimator=type, pairwise=type)
    
class _torch_c(DuckModule):
    # Appease the type checker; ordinarily this binding is inserted by the
    # torch._C module initialization code in C
    __duckmodule__ = 'torch._C'
    __alts__ = dict(_TensorBase=type)
    
class _torch(DuckModule):
    __duckmodule__ = 'torch'
    __alts__ = dict(
        _C=_torch_c, nn=ModuleType('torch.nn'), optim=ModuleType('torch.optim'),
        size=tuple, dtype=type, device=str, Tensor=list,
        IntTensor=list, FloatTensor=list, LongTensor=list, DoubleTensor=list, BoolTensor=list,
        Module=type, Dataset=list, DataLoader=list,
        memory_format=type, reserve_format=type
    )
    
class _torchvision(DuckModule):
    __duckmodule__ = 'torchvision'
    __alts__ = dict(
        datasets=ModuleType('torchvision.datasets'), 
        transforms=ModuleType('torchvision.transforms'), 
        functional=ModuleType('torchvision.transforms.functional'),
    )
    
class _pytorch_lightning(DuckModule):
    __duckmodule__ = 'pytorch_lightning'
    __alts__ = dict(LightningModule=type, LightningDataModule=type)
    
class _torchdyn(DuckModule):
    __duckmodule__ = 'torchdyn'
    __alts__ = dict(
        NeuralODE=type, NeuralSDE=type, ODEFunc=type, SDEFunc=type, Augmenter=type, 
        nn=ModuleType('torchdyn.nn'), utils=ModuleType('torchdyn.utils')
    )
    
class _torchsde(DuckModule):
    __duckmodule__ = 'torchsde'
    __alts__ = dict(SDEIto=type, BrownianInterval=type)
    
class _anndata(DuckModule):
    __duckmodule__ = 'anndata'
    __alts__ = dict(AnnData=list)
    
class _cv2(DuckModule):
    __duckmodule__ = 'cv2'
    __alts__ = dict(VideoCapture=type)

class _PIL(DuckModule):
    __duckmodule__ = 'PIL'
    __alts__ = dict(Image=type)
    
class _graphtools(DuckModule):
    __duckmodule__ = 'graphtools'
    __alts__ = dict(
        Base=type, Data=type, BaseGraph=type, DataGraph=type,
        LandmarkGraph=type, kNNGraph=type, TraditionalGraph=type, MNNGraph=type,
        kNNLandmarkGraph=type, MNNLandmarkGraph=type, TraditionalLandmarkGraph=type
    )

class _phate(DuckModule):
    __duckmodule__ = 'phate'


class _mpl_base(DuckModule):
    __duckmodule__ = 'matplotlib.axes._base'
    
class _mpl_artist(DuckModule):
    __duckmodule__ = 'matplotlib.artist'
    
class _mpl_axes(DuckModule):
    __duckmodule__ = 'matplotlib.axes._axes'
    
class _mpl_patches(DuckModule):
    __duckmodule__ = 'matplotlib.patches'
    
class _matplotlib(DuckModule):
    __duckmodule__ = 'matplotlib'
    __alts__ = dict(
        pyplot=ModuleType('matplotlib.pyplot'),
        Artist=type, Patch=type, Axes=type,
        ColorMap=type, LinearSegmentedColormap=type, 
    )
    
class _seaborn(DuckModule):
    __duckmodule__ = 'seaborn'
    
class _rich(DuckModule):
    __duckmodule__ = 'rich'
    __alts__ = dict(JupyterMixin=type, Console=type, Table=type, Progress=type, Task=type, Text=type, Tree=type)

class _tqdm(DuckModule):
    __duckmodule__ = 'tqdm'
    __alts__ = dict(Comparable=type, tqdm_asyncio=type, tqdm_notebook=type)

class _typer(DuckModule):
    __duckmodule__ = 'typer'
    __alts__ = dict(Typer=type)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_prot.ipynb.

# %% auto 0
__all__ = ['NotProtocol', 'OptProtocol', 'SpecType', 'Duckish', 'Hashable', 'Origin', 'Special', 'NotSpecial',
           'NonStrBytesIterable']

# %% ../nbs/02_prot.ipynb 6
#| export


# %% ../nbs/02_prot.ipynb 8
from types import NoneType
from typing import (Any, Self, TypeGuard, Protocol, Iterable, _SpecialForm, Generic, get_origin,)

# %% ../nbs/02_prot.ipynb 10
#| export

# %% ../nbs/02_prot.ipynb 12
#| export

# %% ../nbs/02_prot.ipynb 14
from .cons import DUCK_SPEC_ATTRS, __HASH__
from .type import T

# %% ../nbs/02_prot.ipynb 16
class NotProtocol(Protocol[T]):
    '''A protocol representing a not being a type'''
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]: 
        return not isinstance(inst, cls)
    
class OptProtocol(Protocol[T]):
    '''A protocol representing an optional type'''
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]: 
        return isinstance(inst, (cls, NoneType))

# %% ../nbs/02_prot.ipynb 18
class SpecType(Protocol[T]):
    '''A protocol representing a duckspec'''
    _attrs: tuple[str, ...] = DUCK_SPEC_ATTRS
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]:
        for attr in cls._attrs:
            if not hasattr(inst, attr): return False
        return True

# %% ../nbs/02_prot.ipynb 19
class Duckish(Protocol[T]):
    '''A protocol representing a something that might be like a duck'''
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]:
        attrs = ('__duck__', 'duck', '_duck', '__ducktypes__')
        if not any(hasattr(inst, attr) for attr in attrs): return False
        return True

# %% ../nbs/02_prot.ipynb 22
class Hashable(Protocol):
    '''A protocol representing a hashable type'''
    def __hash__(self: Self) -> int: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Hashable']:
        return hasattr(__instance, __HASH__)

# %% ../nbs/02_prot.ipynb 24
class Origin(Hashable, Protocol):
    '''A protocol representing origin typing types'''
    def __hash__(self: Self) -> int:...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Origin']:
        return hasattr(__instance, __HASH__) and get_origin(__instance)

# %% ../nbs/02_prot.ipynb 26
class Special(Origin, Protocol):
    '''A protocol representing special typing types'''
    def __hash__(self: Self) -> int:...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Special']:
        origin = super().__instancecheck__(__instance)
        special = isinstance(__instance, _SpecialForm) or __instance in (Generic, Protocol)
        return (origin and special)
    
class NotSpecial(Origin, Protocol):
    '''A protocol representing not special typing types'''
    def __hash__(self: Self) -> int: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['NotSpecial']:
        origin = super().__instancecheck__(__instance)
        special = isinstance(__instance, _SpecialForm) or __instance in (Generic, Protocol)
        return (origin and not special)

# %% ../nbs/02_prot.ipynb 28
class NonStrBytesIterable(Protocol):
    '''A protocol representing non-string, non-bytes iterable objects.'''
    def __iter__(self: Self) -> Iterable: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['NonStrBytesIterable']:
        return isinstance(__instance, Iterable) and not isinstance(__instance, (str, bytes))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_zero.ipynb.

# %% auto 0
__all__ = ['empty']

# %% ../nbs/07_zero.ipynb 6
from itertools import accumulate

# %% ../nbs/07_zero.ipynb 8
from typing import Self, Union

# %% ../nbs/07_zero.ipynb 10
#| export

# %% ../nbs/07_zero.ipynb 12
#| export

# %% ../nbs/07_zero.ipynb 14
#| export


# %% ../nbs/07_zero.ipynb 16
def _shape(shape: tuple = ()) -> tuple:
    '''Ensure that the input is in the form of a tuple.

    Parameters
    ----------
    shape : tuple, optional
        The shape to be converted into a tuple format. If not a tuple, it will be wrapped in one.

    Returns
    -------
    tuple
        The input shape converted into a tuple format.

    '''
    if isinstance(shape, tuple): return shape
    return (shape, )

def _genzeros(shape: tuple = ()) -> list[float]:
    '''Generate a nested list of zeros based on the specified shape.

    Parameters
    ----------
    shape : tuple, optional
        The shape of the nested list to be generated.

    Returns
    -------
    list[float]
        A nested list of zeros with the given shape.
    '''
    if len(shape) == 0: return 0
    if len(shape) == 1: return [0] * shape[0]
    return [_genzeros(shape[1:]) for _ in range(shape[0])]

def _acczeros(shape: tuple = ()) -> list[float]:
    '''Accumulate zeros in a list based on the specified shape.

    Parameters
    ----------
    shape : tuple, optional
        The shape to generate the accumulated zeros for.

    Returns
    -------
    list[float]
        A list of accumulated zeros according to the specified shape.
    '''
    shape = _shape(shape)
    aggfn = lambda x, y: [x] * y
    accum = list(accumulate(shape[::-1], aggfn, initial=0))
    return accum[-1]

# %% ../nbs/07_zero.ipynb 17
def _format_list(
    data: list, prefix: Union[int, str] = 'array('
) -> str:
    '''Format a list into a string with specified indentation and prefix.

    Parameters
    ----------
    data : list
        The list to be formatted.
        
    prefix : Union[int, str], optional
        The prefix to be used for formatting. It can be either a string prefix or 
        an integer representing the indentation level.

    Returns
    -------
    str
        The formatted string representation of the list.
    '''
    indent = len(prefix) if isinstance(prefix, str) else prefix     
    catchr, sepchr = '\n', '], '
    lines = str(data).split(sepchr)
    
    first = lines[0] if len(lines) > 0 else ''
    numbk = first.count('[')
    
    for i in range(len(lines)):
        isfirst, islast = i == 0, i == len(lines) - 1
        # re-add the split character
        if not islast: lines[i] += sepchr
        
        # first line is indented by the characters of `prefix`
        if isfirst: continue
        
        # current number of brackets
        bkcnt = lines[i].count('[')
        # how many brackets are missing (the indent position)
        index = (numbk - bkcnt) + indent
        lines[i] = ' ' * (index)  + lines[i]
        
        # check if next line is last of current level
        if (not islast and bkcnt == 1 and not lines[i + 1].count('[') == 1):
            lines[i] += '\n'
        
    datastr = catchr.join(lines)#.rstrip(sepchr)
    return datastr


# %% ../nbs/07_zero.ipynb 18
class empty(list):
    '''A class representing an empty array with a specified shape.

    Attributes
    ----------
    zeros : list
        A list of zeros according to the specified shape.
        
    shape : tuple
        The shape of the empty array.

    Methods
    -------
    __repr__()
        Return the string representation of the empty array.
    '''
    def __init__(self, shape: tuple= ()):
        shape = _shape(shape)
        zeros = _acczeros(shape)
        super().__init__(zeros)
        self.zeros = zeros
        self.shape = shape
        
    @property
    def ndim(self: Self) -> int:
        '''The number of dimensions of the empty array.'''
        return len(self.shape)
    
    def __repr__(self) -> str: 
        return f'array({_format_list(self.zeros)})'



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/13_duck.ipynb.

# %% auto 0
__all__ = ['moduleloader', 'modulespec', 'iterable', 'itertype', 'moduletype', 'wrappertype', 'builtinfunctype', 'trackbacktype',
           'functiontype', 'mapping', 'dictkeys', 'dictvalues', 'dictitems', 'sig', 'param', 'paramkind', 'logger',
           'ichain', 'iprod', 'ziplong', 'boolic', 'bool01', 'bools', 'intq', 'ints', 'floatq', 'floats', 'real',
           'realq', 'reals', 'num', 'numq', 'nums', 'numstr', 'none', 'nan', 'nanq', 'nans', 'pattern', 'rematch',
           'strtmpl', 'nilstr', 'notnilstr', 'strbytes', 'notstrbytes', 'strq', 'strs', 'ospathlike', 'plibpath',
           'posixpath', 'pathlike', 'artist', 'colormap', 'lcolormap', 'lscolormap', 'mplpatch', 'mplaxbase', 'mplaxes',
           'nparray', 'npmatrix', 'recarray', 'npempty', 'npdatasource', 'sparray', 'spmatrix', 'cscarray', 'cscmatrix',
           'cooarray', 'coomatrix', 'csrarray', 'csrmatrix', 'diaarray', 'diamatrix', 'dokarray', 'dokmatrix',
           'lilarray', 'lilmatrix', 'bsrarray', 'bsrmatrix', 'sparsearray', 'sparsematrix', 'sparse', 'array', 'matrix',
           'cv2vidcap', 'pilimage', 'vidcap', 'image', 'groupkey', 'pddelegate', 'pdcataccessor', 'pdindex',
           'pdcatindex', 'pdseries', 'boolindex', 'catindex', 'indextype', 'indexlike', 'seriestype', 'serieslike',
           'pddataframe', 'dataframe', 'memfmt', 'presfmt', 'ptsize', 'ptdevice', 'pttensor', 'inttensor', 'booltensor',
           'floattensor', 'longtensor', 'doubletensor', 'nantensor', 'tensor', 'ptmodule', 'plmodule', 'nnmodule',
           'ptdataset', 'pldatamodule', 'ptdataloader', 'neuralode', 'neuralsde', 'neuraldynamics', 'augmenter',
           'odefunc', 'sdefunc', 'sdeito', 'brownianintervial', 'dtype', 'device', 'size', 'datasource', 'adata',
           'adatas', 'graphtype', 'datatype', 'shaped', 'dimmed', 'dim1', 'dim2', 'dim3', 'dim4', 'richtree',
           'richtext', 'richcons', 'richprog', 'richtask', 'tqdmauto', 'typerapp']

# %% ../nbs/13_duck.ipynb 6
#| export


# %% ../nbs/13_duck.ipynb 8
from typing import Iterable

# %% ../nbs/13_duck.ipynb 10
#| export

# %% ../nbs/13_duck.ipynb 12
#| export

# %% ../nbs/13_duck.ipynb 14
from .base import Duck
from . import mods
from . import spec
from .enum import Species

# %% ../nbs/13_duck.ipynb 17
class moduleloader(Duck):
    '''`importlib.Loader`'''
    __ducktype__ = (spec.il_moduleloader.duck, )
    
class modulespec(Duck):
    '''`importlib.machinery.ModuleSpec`'''
    __ducktype__ = (spec.il_modulespec.duck, )

# %% ../nbs/13_duck.ipynb 19
class iterable(Duck):
    '''`typings.Iterable`, `collections.abc.Iterable`'''
    __ducktype__ = (Iterable, spec.c_iterable.duck, )
    
class itertype(Duck):
    '''iterable type (i.e. `Iterable`, `list`, or `tuple`)'''
    __ducktype__ = (iterable, list, tuple)

# %% ../nbs/13_duck.ipynb 21
class moduletype(Duck):
    '''`types.ModuleType`'''
    __ducktype__ = (spec.t_moduletype.duck, )
    
class wrappertype(Duck):
    '''`types.MethodWrapperType`'''
    __ducktype__ = (spec.t_methodwrappertype.duck, )
    
class builtinfunctype(Duck):
    '''`types.BuiltinFunctionType`'''
    __ducktype__ = (spec.t_builtinfunctiontype.duck, )
    
class trackbacktype(Duck):
    '''`types.TracebackType`'''
    __ducktype__ = (spec.t_tracebacktype.duck, )
    
class functiontype(Duck):
    '''`types.FunctionType`'''
    __ducktype__ = (spec.t_functiontype.duck, )
    
class mapping(Duck):
    '''`collections.Mapping`'''
    __ducktype__ = (spec.c_mapping.duck, )

# %% ../nbs/13_duck.ipynb 23
class dictkeys(Duck):
    '''`_collections_abc.dict_keys`'''
    __ducktype__ = (spec.c_dict_keys.duck, )
    
class dictvalues(Duck):
    '''`_collections_abc.dict_values`'''
    __ducktype__ = (spec.c_dict_values.duck, )
    
class dictitems(Duck):
    '''`_collections_abc.dict_items`'''
    __ducktype__ = (spec.c_dict_items.duck, )

# %% ../nbs/13_duck.ipynb 25
class sig(Duck):
    '''`inspect.Signature`'''
    __ducktype__ = (spec.i_sig.duck, )
    
class param(Duck):
    '''`inspect.Parameter`'''
    __ducktype__ = (spec.i_param.duck, )
    
class paramkind(Duck):
    '''`inspect._ParameterKind`'''
    __ducktype__ = (spec.i_paramkind.duck, )

# %% ../nbs/13_duck.ipynb 27
class logger(Duck):
    '''`logging.Logger`'''
    __ducktype__ = (spec.l_logger.duck, )

# %% ../nbs/13_duck.ipynb 29
class ichain(Duck):
    '''`itertools.chain`'''
    __ducktype__ = (spec.it_chain.duck, )
    
class iprod(Duck):
    '''`itertools.product`'''
    __ducktype__ = (spec.it_prod.duck, )
    
class ziplong(Duck):
    '''`itertools.zip_longest`'''
    __ducktype__ = (spec.it_ziplong.duck, )

# %% ../nbs/13_duck.ipynb 31
class boolic(Duck): 
    '''`bool`, `np.bool_`'''
    __ducktype__ = (bool, spec.np_bool.duck)

class bool01(Duck):
    '''`bool`, `np.bool_`, `0`, `1`'''
    __ducktype__ = (boolic, 0, 1)
    
class bools(Duck):
    '''list of bool01'''
    __ducktype__ = (itertype, list, tuple)
    __expected__ = dict(dtype=bool01)

# %% ../nbs/13_duck.ipynb 33
class intq(Duck):
    '''`int` or None'''
    __ducktype__ = (int, )
    __species__ = Species.OPT
    
class ints(Duck):
    '''iterable of `int`s'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=int)

# %% ../nbs/13_duck.ipynb 34
class floatq(Duck):
    '''`float` or None'''
    __ducktype__ = (float, )
    __species__ = Species.OPT
    
class floats(Duck):
    '''iterable of `float`s'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=float)

# %% ../nbs/13_duck.ipynb 35
class real(Duck):
    '''R (i.e. `int` or `float`)'''
    __ducktype__ = (int, float, )
    
class realq(Duck):
    '''`real` or None'''
    __ducktype__ = (int, float, )
    __species__ = Species.OPT
    
class reals(Duck):
    '''iterable of `real`'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=real)

# %% ../nbs/13_duck.ipynb 36
class num(Duck):
    '''`int`, `float`, `complex`, or `Number`'''
    __ducktype__ = (int, float, complex, spec.n_number.duck)
    
class numq(Duck):
    '''`real` or None'''
    __ducktype__ = (num, )
    __species__ = Species.OPT
    
class nums(Duck):
    '''iterable of `num`'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=num)

# %% ../nbs/13_duck.ipynb 38
class numstr(Duck):
    '''`num` or str'''
    __ducktype__ = (num, str,)

# %% ../nbs/13_duck.ipynb 40
class none(Duck): 
    '''`None` or `NoneType`'''
    __ducktype__ = (None, spec.t_nonetype.duck)

# %% ../nbs/13_duck.ipynb 42
class nan(Duck):
    '''`math.nan`, `pd.NA`, `np.nan`, `np.NaN`, or `np.NAN`'''
    __ducktype__ = (
        spec.m_nan.duck, spec.np_nan.duck, spec.np_NaN.duck, spec.np_NAN.duck, 
        spec.pd_na.duck, spec.pd_natype.duck, float('nan')
    )
    
class nanq(Duck):
    '''`nan` or None'''
    __ducktype__ = (nan,)
    __species__ = Species.OPT
    
class nans(Duck):
    '''iterable of `nan`'''
    __ducktype__ = (itertype,)
    __expected__ = dict(dtype=nan)

# %% ../nbs/13_duck.ipynb 44
class pattern(Duck): 
    '''`re.Pattern`'''
    __ducktype__ = (spec.re_pattern.duck, )

class rematch(Duck): 
    '''`re.Match`'''
    __ducktype__ = (spec.re_match.duck, )

class strtmpl(Duck): 
    '''`string.Template`'''
    __ducktype__ = (spec.s_template.duck, )

# %% ../nbs/13_duck.ipynb 46
class nilstr(Duck):
    '''`''` (the nil string)'''
    __ducktype__ = ('', )
    __guard__ = lambda x: x == ''
    
class notnilstr(nilstr): 
    '''not `nilstr` (not `''`, the nil string)'''
    __species__ = Species.NOT

# %% ../nbs/13_duck.ipynb 47
class strbytes(Duck):
    '''str or bytes'''
    __ducktype__ = (str, bytes)
    
class notstrbytes(strbytes): 
    '''not (str or bytes)'''
    __species__ = Species.NOT

# %% ../nbs/13_duck.ipynb 48
class strq(Duck): 
    '''`str` or None'''
    __ducktype__ = (str, )
    __species__ = Species.OPT

class strs(Duck): 
    '''iterable of `str`'''
    __ducktype__ = (itertype, notstrbytes, )
    __expected__ = dict(dtype=str)

# %% ../nbs/13_duck.ipynb 50
class ospathlike(Duck):
    '''`os.PathLike`'''
    __ducktype__ = (spec.os_pathlike.duck, )

class plibpath(Duck):
    '''`pathlib.Path`'''
    __ducktype__ = (spec.pl_path.duck, )

class posixpath(Duck):
    '''`pathlib.PosixPath`'''
    __ducktype__ = (spec.pl_posixpath.duck, )

# %% ../nbs/13_duck.ipynb 52
class pathlike(Duck):
    '''`os.PathLike`, `pathlib.Path`, `pathlib.PosixPath`'''
    __ducktype__ = (str, bytes, spec.os_pathlike.duck, spec.pl_path.duck, spec.pl_posixpath.duck)    

# %% ../nbs/13_duck.ipynb 54
class _mpl_base(Duck):
    '''`matplotlib.axes._base`''';
    __ducktype__ = (spec._mpl_base.duck, )
    
class _mpl_axes(Duck):
    '''`matplotlib.axes._axes`''';
    __ducktype__ = (spec._mpl_axes.duck, )
    

class artist(Duck): 
    '''`matplotlib.artist.Artist`'''
    __ducktype__ = (spec.mpl_artist.duck, )
    
class colormap(Duck): 
    '''`matplotlib.colors.Colormap`'''
    __ducktype__ = (spec.mpl_colormap.duck, )

class lcolormap(Duck): 
    '''`matplotlib.colors.ListedColormap`'''
    __ducktype__ = (spec.mpl_lcolormap.duck, )
    
class lscolormap(Duck): 
    '''`matplotlib.colors.LinearSegmentedColormap`'''
    __ducktype__ = (spec.mpl_lscolormap.duck, )
    
class mplpatch(Duck): 
    '''`matplotlib.patches.Patch`'''
    __ducktype__ = (spec.mpl_patch.duck, )

class mplaxbase(Duck): 
    '''`matplotlib.axes._base._AxesBase`'''
    __ducktype__ = (spec.mpl_axbase.duck, )

class mplaxes(Duck): 
    '''`matplotlib.axes._axes.Axes`'''
    __ducktype__ = (spec.mpl_axes.duck, )
    

# %% ../nbs/13_duck.ipynb 56
class nparray(Duck):
    '''`numpy.ndarray`''';
    __ducktype__ = (spec.np_array.duck, )
    
class npmatrix(Duck):
    '''`numpy.matrix`''';
    __ducktype__ = (spec.np_matrix.duck, )

class recarray(Duck):
    '''`numpy.recarray`''';
    __ducktype__ = (spec.np_recarray.duck, )
    
class npempty(Duck):
    '''`numpy.empty`''';
    __ducktype__ = (spec.np_empty.duck, )

# %% ../nbs/13_duck.ipynb 57
class npdatasource(Duck): 
    '''`numpy.DataSource`''';
    __ducktype__ = (spec.np_datasource.duck, )

# %% ../nbs/13_duck.ipynb 59
class sparray(Duck): 
    '''`scipy.sparse.aparray`''';
    __ducktype__ = (spec.sp_sparray.duck, )

class spmatrix(Duck): 
    '''`scipy.sparse.spmatrix`''';
    __ducktype__ = (spec.sp_spmatrix.duck, )
    

# %% ../nbs/13_duck.ipynb 61
class cscarray(Duck):
    '''`scipy.sparse.csc_array`''';
    __ducktype__ = (spec.sp_csc_array.duck, )
    
class cscmatrix(Duck):
    '''`scipy.sparse.csc_matrix`''';
    __ducktype__ = (spec.sp_csc_matrix.duck, )

class cooarray(Duck):
    '''`scipy.sparse.coo_array`''';
    __ducktype__ = (spec.sp_coo_array.duck, )

class coomatrix(Duck):
    '''`scipy.sparse.coo_matrix`''';
    __ducktype__ = (spec.sp_coo_matrix.duck, )

class csrarray(Duck):
    '''`scipy.sparse.csr_array`''';
    __ducktype__ = (spec.sp_csr_array.duck, )

class csrmatrix(Duck):
    '''`scipy.sparse.csr_matrix`''';
    __ducktype__ = (spec.sp_csr_matrix.duck, )

class diaarray(Duck):
    '''`scipy.sparse.dia_array`''';
    __ducktype__ = (spec.sp_dia_array.duck, )

class diamatrix(Duck):
    '''`scipy.sparse.dia_matrix`''';
    __ducktype__ = (spec.sp_dia_matrix.duck, )

class dokarray(Duck):
    '''`scipy.sparse.dok_array`''';
    __ducktype__ = (spec.sp_dok_array.duck, )

class dokmatrix(Duck):
    '''`scipy.sparse.dok_matrix`''';
    __ducktype__ = (spec.sp_dok_matrix.duck, )

class lilarray(Duck):
    '''`scipy.sparse.lil_array`''';
    __ducktype__ = (spec.sp_lil_array.duck, )

class lilmatrix(Duck):
    '''`scipy.sparse.lil_matrix`''';
    __ducktype__ = (spec.sp_lil_matrix.duck, )

class bsrarray(Duck):
    '''`scipy.sparse.bsr_array`''';
    __ducktype__ = (spec.sp_bsr_array.duck, )

class bsrmatrix(Duck):
    '''`scipy.sparse.bsr_matrix`''';
    __ducktype__ = (spec.sp_bsr_matrix.duck, )

# %% ../nbs/13_duck.ipynb 63
class sparsearray(Duck): 
    '''sparse array (e.g. 
    `scipy.sparse.sparray`, `scipy.sparse.cscarray`, `scipy.sparse.cooarray`, 
    `scipy.sparse.csrarray`, `scipy.sparse.diaarray`, `scipy.sparse.dokarray`, 
    `scipy.sparse.lilarray`, or `scipy.sparse.bsrarray`)'''
    __ducktype__ = (sparray, cscarray, cooarray, csrarray, diaarray, dokarray, lilarray, bsrarray)
    
    
class sparsematrix(Duck): 
    '''sparse type (e.g.  
    `scipy.sparse.spmatrix`, `scipy.sparse.cscmatrix`, `scipy.sparse.coomatrix`, 
    `scipy.sparse.csrmatrix`, `scipy.sparse.diamatrix`, `scipy.sparse.dokmatrix`, 
    `scipy.sparse.lilmatrix`, or `scipy.sparse.bsrmatrix`)'''
    __ducktype__ = (spmatrix, cscmatrix, coomatrix, csrmatrix, diamatrix, dokmatrix, lilmatrix, bsrmatrix)


# %% ../nbs/13_duck.ipynb 65
class sparse(Duck): 
    '''sparse type (e.g. `sparsearray` or `sparsematrix`)'''
    __ducktype__ = (sparray, sparsematrix)
    
class array(Duck): 
    '''array type (e.g. `np.array`, `np.recarray`, `sp.sparray`, `itertype`, or `list`)'''
    __ducktype__ = (nparray, recarray, sparray, itertype, list)
    
class matrix(Duck): 
    '''matrix type(e.g. `np.matrix`, `sp.matrix`, `itertype`, or `list`)'''
    __ducktype__ = (npmatrix, spmatrix, itertype, list)

# %% ../nbs/13_duck.ipynb 67
class cv2vidcap(Duck): 
    '''`cv2.VideoCapture`''';
    __ducktype__ = (spec.cv_vidcap.duck, )

class pilimage(Duck): 
    '''`PIL.Image.Image`''';
    __ducktype__ = (spec.pil_image.duck, )    

# %% ../nbs/13_duck.ipynb 69
class vidcap(Duck): 
    '''`cv2.VideoCapture`''';
    __ducktype__ = (spec.cv_vidcap.duck, )
    
class image(Duck): 
    '''`PIL.Image.Image`''';
    __ducktype__ = (spec.pil_image.duck, )

# %% ../nbs/13_duck.ipynb 71
class groupkey(Duck): 
    '''groupby key (e.g. `str`, `tuple`, or `numstr`)'''
    __ducktype__ = (str, tuple, numstr)
    
class pddelegate(Duck):
    '''`pandas.core.accessor.PandasDelegate`''';
    __ducktype__ = (object, spec.pd_delegate.duck, )
    
class pdcataccessor(Duck):
    '''`pandas.core.arrays.categorical.CategoricalAccessor`''';
    __ducktype__ = (object, spec.pd_cataccessor.duck, )

# %% ../nbs/13_duck.ipynb 73
class pdindex(Duck):
    '''`pandas.Index`''';
    __ducktype__ = (spec.pd_index.duck, )
    
class pdcatindex(Duck):
    '''`pandas.CategoricalIndex`''';
    __ducktype__ = (spec.pd_catindex.duck, )

# %% ../nbs/13_duck.ipynb 75
class pdseries(Duck):
    '''`pandas.Series`''';
    __ducktype__ = (spec.pd_series.duck, )
    

# %% ../nbs/13_duck.ipynb 77
class boolindex(Duck): 
    '''boolean index'''
    __ducktype__ = (list, pdindex, nparray, bools, )
    __expected__ = dict(dtype=bool01)
    
class catindex(Duck): 
    '''categorical index'''
    __ducktype__ = (pdcatindex, )

# %% ../nbs/13_duck.ipynb 78
class indextype(Duck): 
    '''index types(e.g. `pd.Index`, `pd.CategoricalIndex`, `np.ndarray`, or `boolindex`)'''
    __ducktype__ = (pdindex, pdcatindex, boolindex, nparray)
    
    
class indexlike(Duck): 
    '''index-like objects (e.g. `list`, `tuple`,`bools`, `nums`, 
    `itertype`, `pd.Index`, `pd.CategoricalIndex`, `boolindex`, or `np.array`)'''
    __ducktype__ = (list, tuple, bools, nums, itertype, pdindex, pdcatindex, boolindex, nparray)

# %% ../nbs/13_duck.ipynb 80
class seriestype(Duck): 
    '''series type (e.g. `pd.Series`, `np.array`, `np.recarray`, `list`, or `indexlike`)'''
    __ducktype__ = (pdseries, nparray, recarray, list, indexlike)
    

class serieslike(Duck): 
    '''series-like type'''
    __ducktype__ = (pdseries, nparray, recarray, list, indexlike)

# %% ../nbs/13_duck.ipynb 82
class pddataframe(Duck): 
    '''`pandas.DataFrame`''';
    __ducktype__ = (spec.pd_dataframe.duck, )

# %% ../nbs/13_duck.ipynb 84
class dataframe(Duck): 
    '''`pandas.DataFrame`''';
    __ducktype__ = (spec.pd_dataframe.duck, )

# %% ../nbs/13_duck.ipynb 87
class memfmt(Duck): 
    '''`torch.memory_format`''';
    __ducktype__ = (spec.pt_memoryformat.duck, )

class presfmt(Duck): 
    '''`torch.preserve_format`''';
    __ducktype__ = (spec.pt_preserveformat.duck, )


# %% ../nbs/13_duck.ipynb 89
class ptsize(Duck): 
    '''`torch.Size`'''
    __ducktype__ = (spec.pt_size.duck, iterable, tuple, )

# %% ../nbs/13_duck.ipynb 91
class ptdevice(Duck): 
    '''`torch.device`'''
    __ducktype__ = (spec.pt_device.duck, iterable, tuple, )


# %% ../nbs/13_duck.ipynb 93
class pttensor(Duck): 
    '''`torch.Tensor`''';
    __ducktype__ = (spec.pt_tensor.duck, )
    
class inttensor(Duck): 
    '''`torch.IntTensor`''';
    __ducktype__ = (spec.pt_inttensor.duck, )
    
class booltensor(Duck): 
    '''`torch.BoolTensor`''';
    __ducktype__ = (spec.pt_booltensor.duck, )
    
class floattensor(Duck): 
    '''`torch.FloatTensor`''';
    __ducktype__ = (spec.pt_floattensor.duck, )
    
class longtensor(Duck): 
    '''`torch.LongTensor`''';
    __ducktype__ = (spec.pt_longtensor.duck, )
    
class doubletensor(Duck): 
    '''`torch.DoubleTensor`''';
    __ducktype__ = (spec.pt_doubletensor.duck, )

# %% ../nbs/13_duck.ipynb 95
class nantensor(Duck): 
    '''tensor of `nan`s'''
    __ducktype__ = (pttensor, nans, )
    __expected__ = dict(dtype=nan)

# %% ../nbs/13_duck.ipynb 97
class tensor(Duck): 
    '''tensor type (e.g. `torch.tensor`, `torch.IntTensor`, `torch.BoolTensor`, or `torch.FloatTensor`)'''
    __ducktype__ = (pttensor, inttensor, booltensor, floattensor, longtensor, doubletensor, )

# %% ../nbs/13_duck.ipynb 99
class ptmodule(Duck): 
    '''`torch.nn.Module`'''
    __ducktype__ = (spec.pt_module.duck, )

class plmodule(Duck): 
    '''`pytorch_lightning.LightningModule`'''
    __ducktype__ = (spec.pl_module.duck, )

# %% ../nbs/13_duck.ipynb 100
class nnmodule(Duck): 
    '''module type (e.g. `torch.nn.Module` or `pytorch_lightning.LightningModule`)'''
    __ducktype__ = (ptmodule, plmodule, )

# %% ../nbs/13_duck.ipynb 102
class ptdataset(Duck): 
    '''`torch.utils.data.Dataset`'''
    __ducktype__ = (spec.pt_dataset.duck, )

class pldatamodule(Duck): 
    '''`pytorch_lightning.LightningDataModule`'''
    __ducktype__ = (spec.pl_datamodule.duck, )

class ptdataloader(Duck): 
    '''`torch.utils.data.DataLoader`'''
    __ducktype__ = (spec.pt_dataloader.duck, )


# %% ../nbs/13_duck.ipynb 104
class neuralode(Duck): 
    '''torchdyn neural ODE'''
    __ducktype__ = (spec.td_neuralode.duck, )

class neuralsde(Duck): 
    '''torchdyn neural SDE'''
    __ducktype__ = (spec.td_neuralsde.duck, )

class neuraldynamics(Duck): 
    '''neuralode or neuralsde'''
    __ducktype__ = (neuralode, neuralsde, )


# %% ../nbs/13_duck.ipynb 105
class augmenter(Duck): 
    '''`torchdyn.nn.augmenter`''';
    __ducktype__ = (spec.td_augmenter.duck, )

# %% ../nbs/13_duck.ipynb 106
class odefunc(Duck): 
    '''`torchdyn.ODEFunc`'''
    __ducktype__ = (spec.td_odefunc.duck, )

class sdefunc(Duck): 
    '''`torchdyn.SDEFunc`'''
    __ducktype__ = (spec.td_sdefunc.duck, )


# %% ../nbs/13_duck.ipynb 108
class sdeito(Duck): 
    '''`torchsde.ito`'''
    __ducktype__ = (spec.ts_sdeito.duck, )

class brownianintervial(Duck): 
    '''`torchsde.BrownianInterval`'''
    __ducktype__ = (spec.ts_brownianinterval.duck, )


# %% ../nbs/13_duck.ipynb 110
class dtype(Duck): 
    '''Type of data (e.g. `np.dtype` or `torch.dtype`)'''
    __ducktype__ = (type, spec.np_type.duck, spec.pt_type.duck, )

# %% ../nbs/13_duck.ipynb 112
class device(Duck): 
    '''`torch.device`'''
    __ducktype__ = (spec.pt_device.duck, )

# %% ../nbs/13_duck.ipynb 114
class size(Duck): 
    '''size type (e.g. `tuple`, `torch.Size`, `itertype`)'''
    __ducktype__ = (tuple, ptsize, itertype, )
    __expected__ = dict(dtype=int)

# %% ../nbs/13_duck.ipynb 116
class datasource(Duck): 
    '''`numpy.DataSource`''';
    __ducktype__ = (spec.np_datasource.duck, )   

# %% ../nbs/13_duck.ipynb 118
class adata(Duck): 
    '''`anndata.AnnData`''';
    __ducktype__ = (spec.ad_adata.duck, )

class adatas(Duck): 
    '''`anndata.AnnData`''';
    __ducktype__ = (itertype, )
    __expected__ = dict(dtype=adata)

# %% ../nbs/13_duck.ipynb 120
class graphtype(Duck): 
    '''graph type (e.g. `BaseGraph`, `DataGraph`, `LandmarkGraph`, `kNNGraph`, `TraditionalGraph`, 
    `MNNGraph`, `kNNLandmarkGraph`, `TraditionalLandmarkGraph`, or `MNNLandmarkGraph`)'''
    __ducktype__ = (
        spec.gt_bgraph.duck, spec.gt_dgraph.duck, spec.gt_lgraph.duck, 
        spec.gt_kgraph.duck, spec.gt_tgraph.duck, spec.gt_mgraph.duck, 
        spec.gt_klgraph.duck, spec.gt_tlgraph.duck, spec.gt_mlgraph.duck
    )

# %% ../nbs/13_duck.ipynb 122
class datatype(Duck): 
    '''data type (e.g.  `np.array`, `np.matrix`, `sp.array`, `sp.matrix`,  `np.recarray`, or `anndata`)'''
    __ducktype__ = (array, matrix, adata, )


# %% ../nbs/13_duck.ipynb 123
class shaped(Duck): 
    '''shaped `datatype`'''
    __ducktype__ = (datatype, indextype, seriestype, dataframe, tensor, )
    __required__ = ('shape', )

# %% ../nbs/13_duck.ipynb 125
class dimmed(Duck): 
    '''`np.ndarray` with shape `(i, ...)`'''
    __ducktype__ = (list, nparray, list, itertype, )
    __required__ = ('ndim', )

class dim1(dimmed): 
    '''`np.ndarray` with shape `(i, )`'''
    __ducktype__ = (dimmed, )
    __expected__ = dict(ndim=1)
    

class dim2(dimmed): 
    '''`np.ndarray` with shape `(i, j, )`'''
    __ducktype__ = (dimmed, )
    __expected__ = dict(ndim=2)
    

class dim3(dimmed): 
    '''`np.ndarray` with shape `(i, j, k, )`'''
    __ducktype__ = (dimmed, )
    __expected__ = dict(ndim=3)

class dim4(dimmed): 
    '''`np.ndarray` with shape `(i, j, k, l, )`'''
    __ducktype__ = (dimmed, )
    __expected__ = dict(ndim=4)

# %% ../nbs/13_duck.ipynb 127
class richtree(Duck):
    '''`rich.tree.Tree`'''
    __ducktype__ = (spec.r_richtree.duck, )
    
class richtext(Duck):
    '''`rich.text.Text`'''
    __ducktype__ = (spec.r_richtext.duck, )
    
class richcons(Duck):
    '''`rich.console.Console`'''
    __ducktype__ = (spec.r_richcons.duck, )
    
class richprog(Duck):
    '''`rich.progress.Progress`'''
    __ducktype__ = (spec.r_richprog.duck, )
    
class richtask(Duck):
    '''`rich.progress.Task`'''
    __ducktype__ = (spec.r_richtask.duck, )

# %% ../nbs/13_duck.ipynb 129
class tqdmauto(Duck):
    '''`tqdm.auto.tqdm`'''
    __ducktype__ = (spec.tq_as.duck, )

# %% ../nbs/13_duck.ipynb 131
class typerapp(Duck):
    '''`typer.Typer`'''
    __ducktype__ = (spec.t_typer.duck, )
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_errs.ipynb.

# %% auto 0
__all__ = ['DuckException', 'DuckNameError', 'DuckImportError', 'DuckAttributeError', 'DuckModuleNotFoundError']

# %% ../nbs/04_errs.ipynb 6
#| export


# %% ../nbs/04_errs.ipynb 8
#| export


# %% ../nbs/04_errs.ipynb 10
#| export

# %% ../nbs/04_errs.ipynb 12
#| export

# %% ../nbs/04_errs.ipynb 14
#| export

# %% ../nbs/04_errs.ipynb 16
class DuckException(Exception): ...
    
class DuckNameError(NameError, DuckException): ...
    
class DuckImportError(ImportError, DuckException): ...
    
class DuckAttributeError(AttributeError, DuckException): ...
    
class DuckModuleNotFoundError(ModuleNotFoundError, DuckException): ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/12_spec.ipynb.

# %% auto 0
__all__ = ['il_moduleloader', 'il_modulespec', 'os_pathlike', 'pl_path', 'pl_posixpath', 'i_sig', 'i_param', 'i_paramkind',
           'i_empty', 'l_logger', 'it_chain', 'it_prod', 'it_ziplong', 'c_dict_keys', 'c_dict_values', 'c_dict_items',
           't_moduletype', 't_methodwrappertype', 't_builtinfunctiontype', 't_tracebacktype', 't_functiontype',
           't_nonetype', 'c_mapping', 'c_iterable', 'm_nan', 'np_nan', 'np_NaN', 'np_NAN', 'pd_na', 'pd_natype',
           'n_number', 're_pattern', 're_match', 's_template', 'np_array', 'np_matrix', 'np_recarray', 'np_bool',
           'np_type', 'np_empty', 'np_datasource', 'sp_sparray', 'sp_spmatrix', 'sp_linalg', 'sp_speigs', 'sp_pdist',
           'sp_csc_array', 'sp_csc_matrix', 'sp_coo_array', 'sp_coo_matrix', 'sp_csr_array', 'sp_csr_matrix',
           'sp_dia_array', 'sp_dia_matrix', 'sp_dok_array', 'sp_dok_matrix', 'sp_lil_array', 'sp_lil_matrix',
           'sp_bsr_array', 'sp_bsr_matrix', 'pd_index', 'pd_series', 'pd_dataframe', 'pd_catindex', 'pd_delegate',
           'pd_cataccessor', 'pt_C', 'pt_nan', 'pt_size', 'pt_type', 'pt_device', 'pt_nn', 'pt_module', 'pt_dataset',
           'pt_dataloader', 'pt_tensorbase', 'pt_tensor', 'pt_inttensor', 'pt_floattensor', 'pt_longtensor',
           'pt_doubletensor', 'pt_booltensor', 'pt_memoryformat', 'pt_preserveformat', 'tv_datasets', 'tv_transforms',
           'tv_functional', 'pl_module', 'pl_datamodule', 'td_odefunc', 'td_sdefunc', 'td_neuralode', 'td_neuralsde',
           'td_augmenter', 'ts_sdeito', 'ts_brownianinterval', 'ad_adata', 'gt_base', 'gt_data', 'gt_bgraph',
           'gt_dgraph', 'gt_lgraph', 'gt_kgraph', 'gt_tgraph', 'gt_mgraph', 'gt_klgraph', 'gt_mlgraph', 'gt_tlgraph',
           'cv_vidcap', 'pil_image', 'mpl', 'mp_pyplot', 'mpl_colormap', 'mpl_lcolormap', 'mpl_lscolormap',
           'mpl_artist', 'mpl_patch', 'mpl_axbase', 'mpl_axes', 'sk_baseestimator', 'sk_pairwise', 'ph_phate',
           'r_jupmix', 'r_richtree', 'r_richtext', 'r_richcons', 'r_richprog', 'r_richtask', 'tq_as', 'tq_nb',
           'tq_auto', 't_typer']

# %% ../nbs/12_spec.ipynb 6
#| export


# %% ../nbs/12_spec.ipynb 8
from abc import ABCMeta
from enum import EnumMeta
from types import ModuleType, FunctionType, BuiltinFunctionType

# %% ../nbs/12_spec.ipynb 10
#| export

# %% ../nbs/12_spec.ipynb 12
#| export

# %% ../nbs/12_spec.ipynb 14
from .base import DuckSpec
from . import mods

# %% ../nbs/12_spec.ipynb 17
il_moduleloader = DuckSpec(asset='importlib.abc.Loader', ducked=mods._importlib, default=type)
'''`importlib.Loader`''';

il_modulespec = DuckSpec(asset='importlib.machinery.ModuleSpec', ducked=mods._importlib, default=type)
'''`importlib.machinery.ModuleSpec`''';

# %% ../nbs/12_spec.ipynb 19
os_pathlike = DuckSpec(asset='os.PathLike', ducked=mods._os, default=str)
'''`os.pathlike`''';

pl_path = DuckSpec(asset='pathlib.Path', ducked=mods._pathlib, default=os_pathlike)
'''`pathlib.Path`''';

pl_posixpath = DuckSpec(asset='pathlib.PosixPath', ducked=mods._pathlib, default=os_pathlike)
'''`pathlib.PosixPath`''';

# %% ../nbs/12_spec.ipynb 21
i_sig = DuckSpec(asset='inspect.Signature', ducked=mods._inspect, default=type)
'''`inspect.Signature`''';

i_param = DuckSpec(asset='inspect.Parameter', ducked=mods._inspect, default=type)
'''`inspect.Parameter`''';

i_paramkind = DuckSpec(asset='inspect._ParameterKind', ducked=mods._inspect, default=EnumMeta)
'''`inspect._ParameterKind`''';

i_empty = DuckSpec(asset='inspect._empty', ducked=mods._inspect, default=type)
'''`inspect.emtpy`''';

# %% ../nbs/12_spec.ipynb 23
l_logger = DuckSpec(asset='logging.Logger', ducked=mods._logging, default=type)
'''`logging.Logger`''';

# %% ../nbs/12_spec.ipynb 25
it_chain = DuckSpec(asset='itertools.chain', ducked=mods._itertools, default=type)
'''`itertools.chain`''';

it_prod = DuckSpec(asset='itertools.product', ducked=mods._itertools, default=type)
'''`itertools.product`''';

it_ziplong = DuckSpec(asset='itertools.zip_longest', ducked=mods._itertools, default=type)
'''`itertools.zip_longest`''';

# %% ../nbs/12_spec.ipynb 28
c_dict_keys = DuckSpec(asset='_collections_abc.dict_keys', ducked=mods._collectionsabc, default=type)
'''`collections.abc.dict_keys`''';

c_dict_values = DuckSpec(asset='_collections_abc.dict_values', ducked=mods._collectionsabc, default=type)
'''`collections.abc.dict_values`''';

c_dict_items = DuckSpec(asset='_collections_abc.dict_items', ducked=mods._collectionsabc, default=type)
'''`collections.abc.dict_items`''';

# %% ../nbs/12_spec.ipynb 31
t_moduletype = DuckSpec(asset='types.ModuleType', ducked=mods._types, default=type)
'''`types.ModuleType`''';

t_methodwrappertype = DuckSpec(asset='types.MethodWrapperType', ducked=mods._types, default=type)
'''`types.MethodWrapperType`''';

t_builtinfunctiontype = DuckSpec(asset='types.BuiltinFunctionType', ducked=mods._types, default=type)
'''`types.BuiltinFunctionType`''';

t_tracebacktype = DuckSpec(asset='types.TracebackType', ducked=mods._types, default=type)
'''`types.TracebackType`''';

t_functiontype = DuckSpec(asset='types.FunctionType', ducked=mods._types, default=type)
'''`types.FunctionType`''';

t_nonetype = DuckSpec(asset='types.NoneType', ducked=mods._types, default=type)
'''`types.NoneType`''';

c_mapping = DuckSpec('collections.abc.Mapping', default=type)
'''`collections.abc.Mapping`''';

c_iterable = DuckSpec('collections.abc.Iterable', default=type)
'''`collections.abc.Iterable`''';

# %% ../nbs/12_spec.ipynb 34
m_nan = DuckSpec(asset='math.nan', ducked=mods._math, default=float('nan'))
'''`math.nan`''';

np_nan = DuckSpec(asset='numpy.nan', ducked=mods._numpy, default=m_nan)
'''`numpy.nan`''';

np_NaN = DuckSpec(asset='numpy.NaN', ducked=mods._numpy, default=m_nan)
'''`numpy.NaN`''';

np_NAN = DuckSpec(asset='numpy.NAN', ducked=mods._numpy, default=m_nan)
'''`numpy.NAN`''';

pd_na = DuckSpec(asset='pandas.NA', ducked=mods._pandas, default=m_nan)
'''`pandas.NA`''';

pd_natype = DuckSpec(asset='pandas._libs.missing.NAType', ducked=mods._pandas, default=type(m_nan))
'''`pandas._libs.missing.NAType`''';

# %% ../nbs/12_spec.ipynb 38
n_number = DuckSpec(asset='numbers.Number', ducked=mods._numbers, default=ABCMeta)
'''`numbers.Number`''';

# %% ../nbs/12_spec.ipynb 40
re_pattern = DuckSpec(asset='re.Pattern', ducked=mods._re, default=type)
'''`re.Pattern`''';

re_match = DuckSpec(asset='re.Match', ducked=mods._re, default=type)
'''`re.Match`''';

s_template = DuckSpec(asset='string.Template', ducked=mods._string, default=type)
'''`string.Template`''';

# %% ../nbs/12_spec.ipynb 43
np_array = DuckSpec(asset='numpy.ndarray', ducked=mods._numpy, default=list)
'''`numpy.ndarray`''';

np_matrix = DuckSpec(asset='numpy.matrix', ducked=mods._numpy, default=list)
'''`numpy.matrix`''';

np_recarray = DuckSpec(asset='numpy.recarray', ducked=mods._numpy, default=list)
'''`numpy.recarray`''';

np_bool = DuckSpec(asset='numpy.bool_', ducked=mods._numpy, default=bool)
'''`numpy.bool_`''';

np_type = DuckSpec(asset='numpy.dtype', ducked=mods._numpy, default=type)
'''`numpy.dtype`''';

np_empty = DuckSpec(asset='numpy.empty', ducked=mods._numpy, default=BuiltinFunctionType)
'''`numpy.empty`''';

np_datasource = DuckSpec(asset='numpy.DataSource', ducked=mods._numpy, default=os_pathlike)
'''`numpy.DataSource`''';

# %% ../nbs/12_spec.ipynb 46
# NOTE: required as scipy.sparse.sparray location depends on version
sp_sparray = DuckSpec(asset='scipy.sparse._arrays._sparray', ducked=mods._scipy, default=list)
sp_sparray = DuckSpec(asset='scipy.sparse.sparray', ducked=mods._scipy, default=(sp_sparray or list))
'''`scipy.sparse.sparray`''';

sp_spmatrix = DuckSpec(asset='scipy.sparse.spmatrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.spmatrix`''';

# %% ../nbs/12_spec.ipynb 47
sp_linalg = DuckSpec(asset='scipy.sparse.linalg', ducked=mods._scipy, default=ModuleType('scipy.sparse.linalg'))
'''`scipy.sparse.linalg`''';

sp_speigs = DuckSpec(asset='scipy.sparse.linalg.eigs', ducked=mods._scipy, default=FunctionType)
'''`scipy.sparse.linalg.eigs`''';

sp_pdist = DuckSpec(asset='scipy.spatial.distance.pdist', ducked=mods._scipy, default=FunctionType)
'''`scipy.spatial.distance.pdist`''';

# %% ../nbs/12_spec.ipynb 49
sp_csc_array = DuckSpec(asset='scipy.sparse.csc_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.csc_array`''';

sp_csc_matrix = DuckSpec(asset='scipy.sparse.csc_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.csc_matrix`''';

sp_coo_array = DuckSpec(asset='scipy.sparse.coo_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.coo_array`''';

sp_coo_matrix = DuckSpec(asset='scipy.sparse.coo_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.coo_matrix`''';

sp_csr_array = DuckSpec(asset='scipy.sparse.csr_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.csr_array`''';

sp_csr_matrix = DuckSpec(asset='scipy.sparse.csr_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.csr_matrix`''';

sp_dia_array = DuckSpec(asset='scipy.sparse.dia_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.dia_array`''';

sp_dia_matrix = DuckSpec(asset='scipy.sparse.dia_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.dia_matrix`''';

sp_dok_array = DuckSpec(asset='scipy.sparse.dok_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.dok_array`''';

sp_dok_matrix = DuckSpec(asset='scipy.sparse.dok_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.dok_matrix`''';

sp_lil_array = DuckSpec(asset='scipy.sparse.lil_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.lil_array`''';

sp_lil_matrix = DuckSpec(asset='scipy.sparse.lil_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.lil_matrix`''';

sp_bsr_array = DuckSpec(asset='scipy.sparse.bsr_array', ducked=mods._scipy, default=list)
'''`scipy.sparse.bsr_array`''';

sp_bsr_matrix = DuckSpec(asset='scipy.sparse.bsr_matrix', ducked=mods._scipy, default=list)
'''`scipy.sparse.bsr_matrix`''';

# %% ../nbs/12_spec.ipynb 51
pd_index = DuckSpec(asset='pandas.Index', ducked=mods._pandas, default=list)
'''`pandas.Index`''';

pd_series = DuckSpec(asset='pandas.Series', ducked=mods._pandas, default=list)
'''`pandas.Series`''';

pd_dataframe = DuckSpec(asset='pandas.DataFrame', ducked=mods._pandas, default=list)
'''`pandas.DataFrame`''';

pd_catindex = DuckSpec(asset='pandas.CategoricalIndex', ducked=mods._pandas, default=list)
'''`pandas.CategoricalIndex`''';

pd_delegate = DuckSpec(asset='pandas.core.accessor.PandasDelegate', ducked=mods._pandas, default=object)
'''`pandas.core.accessor.PandasDelegate`''';

pd_cataccessor = DuckSpec(asset='pandas.core.arrays.categorical.CategoricalAccessor', ducked=mods._pandas, default=object)
'''`pandas.core.arrays.categorical.CategoricalAccessor`''';

# %% ../nbs/12_spec.ipynb 54
pt_C = DuckSpec('torch._C', is_module=True, ducked=mods._torch_c)
'''`torch._C`''';

pt_nan = DuckSpec('torch.nan', ducked=mods._torch_c, default=m_nan)
'''`torch.nan`''';

# %% ../nbs/12_spec.ipynb 56
pt_size = DuckSpec(asset='torch.Size', ducked=mods._torch, default=tuple)
'''`torch.Size`''';

pt_type = DuckSpec(asset='torch.dtype', ducked=mods._torch_c, default=type)
'''`torch.dtype`''';

pt_device = DuckSpec(asset='torch.device', ducked=mods._torch, default=str)
'''`torch.device`''';

# %% ../nbs/12_spec.ipynb 57
pt_nn = DuckSpec(asset='torch.nn', is_module=True, ducked=mods._torch, default=ModuleType('torch.nn'))
'''`torch.nn`''';

# %% ../nbs/12_spec.ipynb 58
pt_module = DuckSpec(asset='torch.nn.Module', ducked=mods._torch, default=type)
'''`torch.nn.Module`''';

pt_dataset = DuckSpec(asset='torch.utils.data.Dataset', ducked=mods._torch, default=list)
'''`torch.utils.data.Dataset`''';

pt_dataloader = DuckSpec(asset='torch.utils.data.DataLoader', ducked=mods._torch, default=list)
'''`torch.utils.data.DataLoader`''';

# %% ../nbs/12_spec.ipynb 59
pt_tensorbase = DuckSpec(asset='torch._C._TensorBase', ducked=mods._torch_c, default=list)
'''`torch._C._TensorBase`''';

pt_tensor = DuckSpec(asset='torch.Tensor', ducked=mods._torch, default=list, bases=(pt_tensorbase, ))
'''`torch.Tensor`''';

pt_inttensor = DuckSpec(asset='torch.IntTensor', ducked=mods._torch, default=list, bases=(pt_tensor, ))
'''`torch.IntTensor`''';

pt_floattensor = DuckSpec(asset='torch.FloatTensor', ducked=mods._torch, default=list, bases=(pt_tensor, ))
'''`torch.FloatTensor`''';

pt_longtensor = DuckSpec(asset='torch.LongTensor', ducked=mods._torch, default=list, bases=(pt_tensor, ))
'''`torch.LongTensor`''';

pt_doubletensor = DuckSpec(asset='torch.DoubleTensor', ducked=mods._torch, default=list, bases=(pt_tensor, ))
'''`torch.DoubleTensor`''';

pt_booltensor = DuckSpec(asset='torch.BoolTensor', ducked=mods._torch, default=list, bases=(pt_tensor, ))
'''`torch.BoolTensor`''';

# %% ../nbs/12_spec.ipynb 60
pt_memoryformat = DuckSpec(asset='torch.memory_format', ducked=mods._torch, default=type)
'''`torch.memory_format`''';

pt_preserveformat = DuckSpec(asset='torch.preserve_format', bases=(pt_memoryformat, ), ducked=mods._torch, default=type)
'''`torch.preserve_format`''';

# %% ../nbs/12_spec.ipynb 63
tv_datasets = DuckSpec(asset='torchvision.datasets', is_module=True, ducked=mods._torchvision, default=ModuleType('torchvision.datasets'))
'''`torchvision.datasets`''';

tv_transforms = DuckSpec(asset='torchvision.transforms', is_module=True, ducked=mods._torchvision, default=ModuleType('torchvision.transforms'))
'''`torchvision.transforms`''';

tv_functional = DuckSpec(asset='torchvision.transforms.functional', is_module=True, ducked=mods._torchvision, default=ModuleType('torchvision.transforms.functional'))
'''`torchvision.transforms.functional`''';

# %% ../nbs/12_spec.ipynb 66
pl_module = DuckSpec(asset='pytorch_lightning.LightningModule', ducked=mods._pytorch_lightning, default=pt_module)
'''`lightning.LightningModule`''';

pl_datamodule = DuckSpec(asset='pytorch_lightning.LightningDataModule', ducked=mods._pytorch_lightning, default=list)
'''`lightning.LightningDataModule`''';

# %% ../nbs/12_spec.ipynb 68
td_odefunc = DuckSpec(asset='torchdyn.core.defunc.ODEFunc', ducked=mods._torchdyn, default=type)
'''`torchdyn.core.defunc.ODEFunc`''';

td_sdefunc = DuckSpec(asset='torchdyn.core.defunc.SDEFunc', ducked=mods._torchdyn, default=type)
'''`torchdyn.core.defunc.SDEFunc`''';

td_neuralode = DuckSpec(asset='torchdyn.core.NeuralODE', ducked=mods._torchdyn, default=pt_module)
'''`torchdyn.core.NeuralODE`''';

td_neuralsde = DuckSpec(asset='torchdyn.core.NeuralSDE', ducked=mods._torchdyn, default=pt_module)
'''`torchdyn.core.NeuralSDE`''';

td_augmenter = DuckSpec(asset='torchdyn.nn.Augmenter', ducked=mods._torchdyn, default=type)
'''`torchdyn.nn.augmenter`''';

# %% ../nbs/12_spec.ipynb 70
ts_sdeito = DuckSpec(asset='torchsde.SDEIto', ducked=mods._torchsde, default=type)
'''`torchsde.ito`''';

ts_brownianinterval = DuckSpec(asset='torchsde.BrownianInterval', ducked=mods._torchsde, default=type)
'''`torchsde.BrownianInterval`''';

# %% ../nbs/12_spec.ipynb 72
ad_adata = DuckSpec(asset='anndata.AnnData', ducked=mods._anndata, default=list)
'''`anndata.AnnData`''';

# %% ../nbs/12_spec.ipynb 75
gt_base = DuckSpec(asset='graphtools.base.Base', ducked=mods._graphtools, default=type)
'''`graphtools.base.Base`''';

gt_data = DuckSpec(asset='graphtools.base.Data', ducked=mods._graphtools, default=type, bases=(gt_base, ))
'''`graphtools.base.Data`''';

gt_bgraph = DuckSpec(asset='graphtools.base.BaseGraph', ducked=mods._graphtools, default=type, bases=(gt_base, ))
'''`graphtools.base.BaseGraph`''';

gt_dgraph = DuckSpec(asset='graphtools.base.DataGraph', ducked=mods._graphtools, default=type, bases=(gt_base, gt_bgraph))
'''`graphtools.base.DataGraph`''';

# %% ../nbs/12_spec.ipynb 76
gt_lgraph = DuckSpec(asset='graphtools.graphs.LandmarkGraph', ducked=mods._graphtools, default=type, bases=(gt_dgraph, ))
'''`graphtools.graphs.LandmarkGraph`''';

gt_kgraph = DuckSpec(asset='graphtools.graphs.kNNGraph', ducked=mods._graphtools, default=type, bases=(gt_dgraph, ))
'''`graphtools.graphs.kNNGraph`''';

gt_tgraph = DuckSpec(asset='graphtools.graphs.TraditionalGraph', ducked=mods._graphtools, default=type, bases=(gt_dgraph, ))
'''`graphtools.graphs.TraditionalGraph`''';

gt_mgraph = DuckSpec(asset='graphtools.graphs.MNNGraph', ducked=mods._graphtools, default=type, bases=(gt_dgraph, ))
'''`graphtools.graphs.MNNGraph`''';

# %% ../nbs/12_spec.ipynb 77
gt_klgraph = DuckSpec(asset='graphtools.graphs.kNNLandmarkGraph', ducked=mods._graphtools, default=type, bases=(gt_kgraph, gt_lgraph, ))
'''`graphtools.graphs.kNNLandmarkGraph`''';

gt_mlgraph = DuckSpec(asset='graphtools.graphs.MNNLandmarkGraph', ducked=mods._graphtools, default=type, bases=(gt_mgraph, gt_lgraph, ))
'''`graphtools.graphs.MNNLandmarkGraph`''';

gt_tlgraph = DuckSpec(asset='graphtools.graphs.TraditionalLandmarkGraph', ducked=mods._graphtools, default=type, bases=(gt_tgraph, gt_lgraph, ))
'''`graphtools.graphs.TraditionalLandmarkGraph`''';

# %% ../nbs/12_spec.ipynb 80
cv_vidcap = DuckSpec(asset='cv2.VideoCapture', ducked=mods._cv2, default=type)
'''`cv2.VideoCapture`''';

# %% ../nbs/12_spec.ipynb 82
pil_image = DuckSpec(asset='PIL.Image.Image', ducked=mods._PIL, default=type)
'''`PIL.Image.Image`''';

# %% ../nbs/12_spec.ipynb 84
mpl = DuckSpec(asset='matplotlib', is_module=True, ducked=mods._matplotlib, default=ModuleType('matplotlib'))
'''`matplotlib`''';

mp_pyplot = DuckSpec(asset='matplotlib.pyplot', is_module=True, ducked=mods._matplotlib, default=ModuleType('matplotlib.pyplot'))
'''`matplotlib.pyplot`''';

mpl_colormap = DuckSpec(asset='matplotlib.colors.Colormap', ducked=mods._matplotlib, default=type)
'''`matplotlib.colors.Colormap`''';

mpl_lcolormap = DuckSpec(asset='matplotlib.colors.ListedColormap', ducked=mods._matplotlib, default=type, bases=(mpl_colormap, ))
'''`matplotlib.colors.ListedColormap`''';

mpl_lscolormap = DuckSpec(asset='matplotlib.colors.LinearSegmentedColormap', ducked=mods._matplotlib, default=type, bases=(mpl_colormap, ))
'''`matplotlib.colors.LinearSegmentedColormap`''';

_mpl_base = DuckSpec(asset='matplotlib.axes._base', is_module=True, ducked=mods._mpl_base, default=ModuleType('matplotlib.axes._base'))
'''`matplotlib.axes._base`''';

_mpl_axes = DuckSpec(asset='matplotlib.axes._axes', is_module=True, ducked=mods._mpl_axes, default=ModuleType('matplotlib.axes._axes'))
'''`matplotlib.axes._axes`''';

# _mpl_base._AxesBase
mpl_artist = DuckSpec(asset='matplotlib.artist.Artist', ducked=mods._mpl_artist, default=type)
'''`matplotlib.artist.Artist`''';

mpl_patch = DuckSpec(asset='matplotlib.patches.Patch', ducked=mods._mpl_patches, default=type, bases=lambda: (mpl_artist.duck, ), delayed=dict(bases=True))
'''`matplotlib.patches.Patch`''';

mpl_axbase = DuckSpec(asset='matplotlib.axes._base._AxesBase', ducked=mods._mpl_base, default=type, bases=lambda: (mpl_artist.duck, ), delayed=dict(bases=True))
'''`matplotlib.axes._base._AxesBase`''';

mpl_axes = DuckSpec(asset='matplotlib.axes._axes.Axes', ducked=mods._mpl_axes, default=type, bases=lambda: (mpl_axbase.duck, ), delayed=dict(bases=True))
'''`matplotlib.axes._axes.Axes`''';

# %% ../nbs/12_spec.ipynb 87
sk_baseestimator = DuckSpec(asset='sklearn.base.BaseEstimator', ducked=mods._sklearn, default=type)
'''`sklearn.base.BaseEstimator`''';

sk_pairwise = DuckSpec(asset='sklearn.metrics.pairwise', is_module=True, ducked=mods._sklearn, default=ModuleType('sklearn.metrics.pairwise'))
'''`sklearn.metrics.pairwise`''';

# %% ../nbs/12_spec.ipynb 90
ph_phate = DuckSpec(asset='phate.PHATE', ducked=mods._phate, default=sk_baseestimator)
'''`phate.PHATE`''';

# %% ../nbs/12_spec.ipynb 92
r_jupmix = DuckSpec(asset='rich.jupyter.JupyterMixin', ducked=mods._rich, default=type)
'''`rich.jupyter.JupyterMixin`''';

r_richtree = DuckSpec(asset='rich.tree.Tree', ducked=mods._rich, default=type, bases=(r_jupmix, ))
'''`rich.tree.Tree`''';

r_richtext = DuckSpec(asset='rich.text.Text', ducked=mods._rich, default=type, bases=(r_jupmix, ))
'''`rich.text.Text`''';

r_richcons = DuckSpec(asset='rich.console.Console', ducked=mods._rich, default=type, bases=(r_jupmix, ))
'''`rich.console.Console`''';

r_richprog = DuckSpec(asset='rich.progress.Progress', ducked=mods._rich, default=type, bases=(r_jupmix, ))
'''`rich.progress.Progress`''';

r_richtask = DuckSpec(asset='rich.progress.Task', ducked=mods._rich, default=type, bases=(r_jupmix, ))
'''`rich.progress.Task`''';

# %% ../nbs/12_spec.ipynb 94
_tq_com = DuckSpec(asset='tqdm.utils.Comparable', ducked=mods._tqdm, default=type)
'''`tqdm.utils.Comparable`''';

_tq_std = DuckSpec(asset='tqdm.std.tqdm', ducked=mods._tqdm, default=type, bases=(_tq_com, ))
'''`tqdm.std.tqdm`''';

tq_as = DuckSpec(asset='tqdm.asyncio.tqdm_asyncio', ducked=mods._tqdm, default=type, bases=(_tq_std, ))
'''`tqdm.tqdm_asyncio.tqdm_asyncio`''';

tq_nb = DuckSpec(asset='tqdm.notebook.tqdm_notebook', ducked=mods._tqdm, default=type, bases=(_tq_std, ))
'''`tqdm.notebook.tqdm_notebook`''';

tq_auto = DuckSpec(asset='tqdm.auto.tqdm', ducked=mods._tqdm, default=lambda x: x, bases=(tq_nb, tq_as))
'''`tqdm.auto.tqdm`''';

# %% ../nbs/12_spec.ipynb 96
t_typer = DuckSpec(asset='typer.Typer', ducked=mods._typer, default=type)
'''`typer.Typer`''';



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/quac',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/quac',
                'lib_path': 'quac'},
  'syms': { 'quac.base': { 'quac.base.Duck': ('base.html#duck', 'quac/base.py'),
                           'quac.base.DuckModule': ('base.html#duckmodule', 'quac/base.py'),
                           'quac.base.DuckSpec': ('base.html#duckspec', 'quac/base.py'),
                           'quac.base.DuckSpec.__init__': ('base.html#duckspec.__init__', 'quac/base.py'),
                           'quac.base.DuckSpec.__load__': ('base.html#duckspec.__load__', 'quac/base.py'),
                           'quac.base.DuckSpec.__new__': ('base.html#duckspec.__new__', 'quac/base.py'),
                           'quac.base.DuckSpec.__prep__': ('base.html#duckspec.__prep__', 'quac/base.py'),
                           'quac.base.DuckSpec.__repr__': ('base.html#duckspec.__repr__', 'quac/base.py'),
                           'quac.base.DuckSpec.duck': ('base.html#duckspec.duck', 'quac/base.py'),
                           'quac.base.DuckSpec.load': ('base.html#duckspec.load', 'quac/base.py')},
            'quac.cons': {},
            'quac.duck': { 'quac.duck._mpl_axes': ('duck.html#_mpl_axes', 'quac/duck.py'),
                           'quac.duck._mpl_base': ('duck.html#_mpl_base', 'quac/duck.py'),
                           'quac.duck.adata': ('duck.html#adata', 'quac/duck.py'),
                           'quac.duck.adatas': ('duck.html#adatas', 'quac/duck.py'),
                           'quac.duck.array': ('duck.html#array', 'quac/duck.py'),
                           'quac.duck.artist': ('duck.html#artist', 'quac/duck.py'),
                           'quac.duck.augmenter': ('duck.html#augmenter', 'quac/duck.py'),
                           'quac.duck.bool01': ('duck.html#bool01', 'quac/duck.py'),
                           'quac.duck.boolic': ('duck.html#boolic', 'quac/duck.py'),
                           'quac.duck.boolindex': ('duck.html#boolindex', 'quac/duck.py'),
                           'quac.duck.bools': ('duck.html#bools', 'quac/duck.py'),
                           'quac.duck.booltensor': ('duck.html#booltensor', 'quac/duck.py'),
                           'quac.duck.brownianintervial': ('duck.html#brownianintervial', 'quac/duck.py'),
                           'quac.duck.bsrarray': ('duck.html#bsrarray', 'quac/duck.py'),
                           'quac.duck.bsrmatrix': ('duck.html#bsrmatrix', 'quac/duck.py'),
                           'quac.duck.builtinfunctype': ('duck.html#builtinfunctype', 'quac/duck.py'),
                           'quac.duck.catindex': ('duck.html#catindex', 'quac/duck.py'),
                           'quac.duck.colormap': ('duck.html#colormap', 'quac/duck.py'),
                           'quac.duck.cooarray': ('duck.html#cooarray', 'quac/duck.py'),
                           'quac.duck.coomatrix': ('duck.html#coomatrix', 'quac/duck.py'),
                           'quac.duck.cscarray': ('duck.html#cscarray', 'quac/duck.py'),
                           'quac.duck.cscmatrix': ('duck.html#cscmatrix', 'quac/duck.py'),
                           'quac.duck.csrarray': ('duck.html#csrarray', 'quac/duck.py'),
                           'quac.duck.csrmatrix': ('duck.html#csrmatrix', 'quac/duck.py'),
                           'quac.duck.cv2vidcap': ('duck.html#cv2vidcap', 'quac/duck.py'),
                           'quac.duck.dataframe': ('duck.html#dataframe', 'quac/duck.py'),
                           'quac.duck.datasource': ('duck.html#datasource', 'quac/duck.py'),
                           'quac.duck.datatype': ('duck.html#datatype', 'quac/duck.py'),
                           'quac.duck.device': ('duck.html#device', 'quac/duck.py'),
                           'quac.duck.diaarray': ('duck.html#diaarray', 'quac/duck.py'),
                           'quac.duck.diamatrix': ('duck.html#diamatrix', 'quac/duck.py'),
                           'quac.duck.dictitems': ('duck.html#dictitems', 'quac/duck.py'),
                           'quac.duck.dictkeys': ('duck.html#dictkeys', 'quac/duck.py'),
                           'quac.duck.dictvalues': ('duck.html#dictvalues', 'quac/duck.py'),
                           'quac.duck.dim1': ('duck.html#dim1', 'quac/duck.py'),
                           'quac.duck.dim2': ('duck.html#dim2', 'quac/duck.py'),
                           'quac.duck.dim3': ('duck.html#dim3', 'quac/duck.py'),
                           'quac.duck.dim4': ('duck.html#dim4', 'quac/duck.py'),
                           'quac.duck.dimmed': ('duck.html#dimmed', 'quac/duck.py'),
                           'quac.duck.dokarray': ('duck.html#dokarray', 'quac/duck.py'),
                           'quac.duck.dokmatrix': ('duck.html#dokmatrix', 'quac/duck.py'),
                           'quac.duck.doubletensor': ('duck.html#doubletensor', 'quac/duck.py'),
                           'quac.duck.dtype': ('duck.html#dtype', 'quac/duck.py'),
                           'quac.duck.floatq': ('duck.html#floatq', 'quac/duck.py'),
                           'quac.duck.floats': ('duck.html#floats', 'quac/duck.py'),
                           'quac.duck.floattensor': ('duck.html#floattensor', 'quac/duck.py'),
                           'quac.duck.functiontype': ('duck.html#functiontype', 'quac/duck.py'),
                           'quac.duck.graphtype': ('duck.html#graphtype', 'quac/duck.py'),
                           'quac.duck.groupkey': ('duck.html#groupkey', 'quac/duck.py'),
                           'quac.duck.ichain': ('duck.html#ichain', 'quac/duck.py'),
                           'quac.duck.image': ('duck.html#image', 'quac/duck.py'),
                           'quac.duck.indexlike': ('duck.html#indexlike', 'quac/duck.py'),
                           'quac.duck.indextype': ('duck.html#indextype', 'quac/duck.py'),
                           'quac.duck.intq': ('duck.html#intq', 'quac/duck.py'),
                           'quac.duck.ints': ('duck.html#ints', 'quac/duck.py'),
                           'quac.duck.inttensor': ('duck.html#inttensor', 'quac/duck.py'),
                           'quac.duck.iprod': ('duck.html#iprod', 'quac/duck.py'),
                           'quac.duck.iterable': ('duck.html#iterable', 'quac/duck.py'),
                           'quac.duck.itertype': ('duck.html#itertype', 'quac/duck.py'),
                           'quac.duck.lcolormap': ('duck.html#lcolormap', 'quac/duck.py'),
                           'quac.duck.lilarray': ('duck.html#lilarray', 'quac/duck.py'),
                           'quac.duck.lilmatrix': ('duck.html#lilmatrix', 'quac/duck.py'),
                           'quac.duck.logger': ('duck.html#logger', 'quac/duck.py'),
                           'quac.duck.longtensor': ('duck.html#longtensor', 'quac/duck.py'),
                           'quac.duck.lscolormap': ('duck.html#lscolormap', 'quac/duck.py'),
                           'quac.duck.mapping': ('duck.html#mapping', 'quac/duck.py'),
                           'quac.duck.matrix': ('duck.html#matrix', 'quac/duck.py'),
                           'quac.duck.memfmt': ('duck.html#memfmt', 'quac/duck.py'),
                           'quac.duck.moduleloader': ('duck.html#moduleloader', 'quac/duck.py'),
                           'quac.duck.modulespec': ('duck.html#modulespec', 'quac/duck.py'),
                           'quac.duck.moduletype': ('duck.html#moduletype', 'quac/duck.py'),
                           'quac.duck.mplaxbase': ('duck.html#mplaxbase', 'quac/duck.py'),
                           'quac.duck.mplaxes': ('duck.html#mplaxes', 'quac/duck.py'),
                           'quac.duck.mplpatch': ('duck.html#mplpatch', 'quac/duck.py'),
                           'quac.duck.nan': ('duck.html#nan', 'quac/duck.py'),
                           'quac.duck.nanq': ('duck.html#nanq', 'quac/duck.py'),
                           'quac.duck.nans': ('duck.html#nans', 'quac/duck.py'),
                           'quac.duck.nantensor': ('duck.html#nantensor', 'quac/duck.py'),
                           'quac.duck.neuraldynamics': ('duck.html#neuraldynamics', 'quac/duck.py'),
                           'quac.duck.neuralode': ('duck.html#neuralode', 'quac/duck.py'),
                           'quac.duck.neuralsde': ('duck.html#neuralsde', 'quac/duck.py'),
                           'quac.duck.nilstr': ('duck.html#nilstr', 'quac/duck.py'),
                           'quac.duck.nnmodule': ('duck.html#nnmodule', 'quac/duck.py'),
                           'quac.duck.none': ('duck.html#none', 'quac/duck.py'),
                           'quac.duck.notnilstr': ('duck.html#notnilstr', 'quac/duck.py'),
                           'quac.duck.notstrbytes': ('duck.html#notstrbytes', 'quac/duck.py'),
                           'quac.duck.nparray': ('duck.html#nparray', 'quac/duck.py'),
                           'quac.duck.npdatasource': ('duck.html#npdatasource', 'quac/duck.py'),
                           'quac.duck.npempty': ('duck.html#npempty', 'quac/duck.py'),
                           'quac.duck.npmatrix': ('duck.html#npmatrix', 'quac/duck.py'),
                           'quac.duck.num': ('duck.html#num', 'quac/duck.py'),
                           'quac.duck.numq': ('duck.html#numq', 'quac/duck.py'),
                           'quac.duck.nums': ('duck.html#nums', 'quac/duck.py'),
                           'quac.duck.numstr': ('duck.html#numstr', 'quac/duck.py'),
                           'quac.duck.odefunc': ('duck.html#odefunc', 'quac/duck.py'),
                           'quac.duck.ospathlike': ('duck.html#ospathlike', 'quac/duck.py'),
                           'quac.duck.param': ('duck.html#param', 'quac/duck.py'),
                           'quac.duck.paramkind': ('duck.html#paramkind', 'quac/duck.py'),
                           'quac.duck.pathlike': ('duck.html#pathlike', 'quac/duck.py'),
                           'quac.duck.pattern': ('duck.html#pattern', 'quac/duck.py'),
                           'quac.duck.pdcataccessor': ('duck.html#pdcataccessor', 'quac/duck.py'),
                           'quac.duck.pdcatindex': ('duck.html#pdcatindex', 'quac/duck.py'),
                           'quac.duck.pddataframe': ('duck.html#pddataframe', 'quac/duck.py'),
                           'quac.duck.pddelegate': ('duck.html#pddelegate', 'quac/duck.py'),
                           'quac.duck.pdindex': ('duck.html#pdindex', 'quac/duck.py'),
                           'quac.duck.pdseries': ('duck.html#pdseries', 'quac/duck.py'),
                           'quac.duck.pilimage': ('duck.html#pilimage', 'quac/duck.py'),
                           'quac.duck.pldatamodule': ('duck.html#pldatamodule', 'quac/duck.py'),
                           'quac.duck.plibpath': ('duck.html#plibpath', 'quac/duck.py'),
                           'quac.duck.plmodule': ('duck.html#plmodule', 'quac/duck.py'),
                           'quac.duck.posixpath': ('duck.html#posixpath', 'quac/duck.py'),
                           'quac.duck.presfmt': ('duck.html#presfmt', 'quac/duck.py'),
                           'quac.duck.ptdataloader': ('duck.html#ptdataloader', 'quac/duck.py'),
                           'quac.duck.ptdataset': ('duck.html#ptdataset', 'quac/duck.py'),
                           'quac.duck.ptdevice': ('duck.html#ptdevice', 'quac/duck.py'),
                           'quac.duck.ptmodule': ('duck.html#ptmodule', 'quac/duck.py'),
                           'quac.duck.ptsize': ('duck.html#ptsize', 'quac/duck.py'),
                           'quac.duck.pttensor': ('duck.html#pttensor', 'quac/duck.py'),
                           'quac.duck.real': ('duck.html#real', 'quac/duck.py'),
                           'quac.duck.realq': ('duck.html#realq', 'quac/duck.py'),
                           'quac.duck.reals': ('duck.html#reals', 'quac/duck.py'),
                           'quac.duck.recarray': ('duck.html#recarray', 'quac/duck.py'),
                           'quac.duck.rematch': ('duck.html#rematch', 'quac/duck.py'),
                           'quac.duck.richcons': ('duck.html#richcons', 'quac/duck.py'),
                           'quac.duck.richprog': ('duck.html#richprog', 'quac/duck.py'),
                           'quac.duck.richtask': ('duck.html#richtask', 'quac/duck.py'),
                           'quac.duck.richtext': ('duck.html#richtext', 'quac/duck.py'),
                           'quac.duck.richtree': ('duck.html#richtree', 'quac/duck.py'),
                           'quac.duck.sdefunc': ('duck.html#sdefunc', 'quac/duck.py'),
                           'quac.duck.sdeito': ('duck.html#sdeito', 'quac/duck.py'),
                           'quac.duck.serieslike': ('duck.html#serieslike', 'quac/duck.py'),
                           'quac.duck.seriestype': ('duck.html#seriestype', 'quac/duck.py'),
                           'quac.duck.shaped': ('duck.html#shaped', 'quac/duck.py'),
                           'quac.duck.sig': ('duck.html#sig', 'quac/duck.py'),
                           'quac.duck.size': ('duck.html#size', 'quac/duck.py'),
                           'quac.duck.sparray': ('duck.html#sparray', 'quac/duck.py'),
                           'quac.duck.sparse': ('duck.html#sparse', 'quac/duck.py'),
                           'quac.duck.sparsearray': ('duck.html#sparsearray', 'quac/duck.py'),
                           'quac.duck.sparsematrix': ('duck.html#sparsematrix', 'quac/duck.py'),
                           'quac.duck.spmatrix': ('duck.html#spmatrix', 'quac/duck.py'),
                           'quac.duck.strbytes': ('duck.html#strbytes', 'quac/duck.py'),
                           'quac.duck.strq': ('duck.html#strq', 'quac/duck.py'),
                           'quac.duck.strs': ('duck.html#strs', 'quac/duck.py'),
                           'quac.duck.strtmpl': ('duck.html#strtmpl', 'quac/duck.py'),
                           'quac.duck.tensor': ('duck.html#tensor', 'quac/duck.py'),
                           'quac.duck.tqdmauto': ('duck.html#tqdmauto', 'quac/duck.py'),
                           'quac.duck.trackbacktype': ('duck.html#trackbacktype', 'quac/duck.py'),
                           'quac.duck.typerapp': ('duck.html#typerapp', 'quac/duck.py'),
                           'quac.duck.vidcap': ('duck.html#vidcap', 'quac/duck.py'),
                           'quac.duck.wrappertype': ('duck.html#wrappertype', 'quac/duck.py'),
                           'quac.duck.ziplong': ('duck.html#ziplong', 'quac/duck.py')},
            'quac.enum': { 'quac.enum.LoaderState': ('enum.html#loaderstate', 'quac/enum.py'),
                           'quac.enum.LoaderState._missing_': ('enum.html#loaderstate._missing_', 'quac/enum.py'),
                           'quac.enum.LoaderState.exec': ('enum.html#loaderstate.exec', 'quac/enum.py'),
                           'quac.enum.LoaderState.get': ('enum.html#loaderstate.get', 'quac/enum.py'),
                           'quac.enum.LoaderState.set': ('enum.html#loaderstate.set', 'quac/enum.py'),
                           'quac.enum.LoaderState.spec': ('enum.html#loaderstate.spec', 'quac/enum.py'),
                           'quac.enum.Species': ('enum.html#species', 'quac/enum.py')},
            'quac.errs': { 'quac.errs.DuckAttributeError': ('errs.html#duckattributeerror', 'quac/errs.py'),
                           'quac.errs.DuckException': ('errs.html#duckexception', 'quac/errs.py'),
                           'quac.errs.DuckImportError': ('errs.html#duckimporterror', 'quac/errs.py'),
                           'quac.errs.DuckModuleNotFoundError': ('errs.html#duckmodulenotfounderror', 'quac/errs.py'),
                           'quac.errs.DuckNameError': ('errs.html#ducknameerror', 'quac/errs.py')},
            'quac.grds': { 'quac.grds.allnotspecial': ('grds.html#allnotspecial', 'quac/grds.py'),
                           'quac.grds.getnotspecials': ('grds.html#getnotspecials', 'quac/grds.py'),
                           'quac.grds.hasduck': ('grds.html#hasduck', 'quac/grds.py'),
                           'quac.grds.hashash': ('grds.html#hashash', 'quac/grds.py'),
                           'quac.grds.hasload': ('grds.html#hasload', 'quac/grds.py'),
                           'quac.grds.hasorigin': ('grds.html#hasorigin', 'quac/grds.py'),
                           'quac.grds.isalias': ('grds.html#isalias', 'quac/grds.py'),
                           'quac.grds.isallsame': ('grds.html#isallsame', 'quac/grds.py'),
                           'quac.grds.isargs': ('grds.html#isargs', 'quac/grds.py'),
                           'quac.grds.isbool': ('grds.html#isbool', 'quac/grds.py'),
                           'quac.grds.iscls': ('grds.html#iscls', 'quac/grds.py'),
                           'quac.grds.isclstype': ('grds.html#isclstype', 'quac/grds.py'),
                           'quac.grds.isdict': ('grds.html#isdict', 'quac/grds.py'),
                           'quac.grds.isduckmodule': ('grds.html#isduckmodule', 'quac/grds.py'),
                           'quac.grds.isduckspec': ('grds.html#isduckspec', 'quac/grds.py'),
                           'quac.grds.ishashable': ('grds.html#ishashable', 'quac/grds.py'),
                           'quac.grds.isiter': ('grds.html#isiter', 'quac/grds.py'),
                           'quac.grds.isnan': ('grds.html#isnan', 'quac/grds.py'),
                           'quac.grds.isnone': ('grds.html#isnone', 'quac/grds.py'),
                           'quac.grds.isnotiterstr': ('grds.html#isnotiterstr', 'quac/grds.py'),
                           'quac.grds.isnotspecial': ('grds.html#isnotspecial', 'quac/grds.py'),
                           'quac.grds.isorigin': ('grds.html#isorigin', 'quac/grds.py'),
                           'quac.grds.isspecial': ('grds.html#isspecial', 'quac/grds.py'),
                           'quac.grds.isstr': ('grds.html#isstr', 'quac/grds.py'),
                           'quac.grds.istuple': ('grds.html#istuple', 'quac/grds.py'),
                           'quac.grds.istype': ('grds.html#istype', 'quac/grds.py'),
                           'quac.grds.istyping': ('grds.html#istyping', 'quac/grds.py'),
                           'quac.grds.notcls': ('grds.html#notcls', 'quac/grds.py'),
                           'quac.grds.notnone': ('grds.html#notnone', 'quac/grds.py'),
                           'quac.grds.notspecialsorigins': ('grds.html#notspecialsorigins', 'quac/grds.py')},
            'quac.inst': {},
            'quac.meta': { 'quac.meta.ABCDuck': ('meta.html#abcduck', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__alias__': ('meta.html#abcduck.__alias__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__call__': ('meta.html#abcduck.__call__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__contains__': ('meta.html#abcduck.__contains__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__eq__': ('meta.html#abcduck.__eq__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__guard__': ('meta.html#abcduck.__guard__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__guardq__': ('meta.html#abcduck.__guardq__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__instancecheck__': ('meta.html#abcduck.__instancecheck__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__invert__': ('meta.html#abcduck.__invert__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__membercheck__': ('meta.html#abcduck.__membercheck__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__memberguard__': ('meta.html#abcduck.__memberguard__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__notinstance__': ('meta.html#abcduck.__notinstance__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__optinstance__': ('meta.html#abcduck.__optinstance__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__repr__': ('meta.html#abcduck.__repr__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__subclasscheck__': ('meta.html#abcduck.__subclasscheck__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__types__': ('meta.html#abcduck.__types__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__typesq__': ('meta.html#abcduck.__typesq__', 'quac/meta.py'),
                           'quac.meta.ABCDuck.__usetypes__': ('meta.html#abcduck.__usetypes__', 'quac/meta.py'),
                           'quac.meta.ABCSpecies': ('meta.html#abcspecies', 'quac/meta.py'),
                           'quac.meta.ABCSpecies.__migrate__': ('meta.html#abcspecies.__migrate__', 'quac/meta.py'),
                           'quac.meta.ABCSpecies.__not__': ('meta.html#abcspecies.__not__', 'quac/meta.py'),
                           'quac.meta.ABCSpecies.__opt__': ('meta.html#abcspecies.__opt__', 'quac/meta.py'),
                           'quac.meta.ABCSpecies.__reg__': ('meta.html#abcspecies.__reg__', 'quac/meta.py'),
                           'quac.meta.DuckModuleMeta': ('meta.html#duckmodulemeta', 'quac/meta.py'),
                           'quac.meta.DuckModuleMeta.__getattr__': ('meta.html#duckmodulemeta.__getattr__', 'quac/meta.py'),
                           'quac.meta.DuckModuleMeta.__instancecheck__': ('meta.html#duckmodulemeta.__instancecheck__', 'quac/meta.py')},
            'quac.mods': { 'quac.mods._PIL': ('mods.html#_pil', 'quac/mods.py'),
                           'quac.mods._anndata': ('mods.html#_anndata', 'quac/mods.py'),
                           'quac.mods._collectionsabc': ('mods.html#_collectionsabc', 'quac/mods.py'),
                           'quac.mods._cv2': ('mods.html#_cv2', 'quac/mods.py'),
                           'quac.mods._graphtools': ('mods.html#_graphtools', 'quac/mods.py'),
                           'quac.mods._importlib': ('mods.html#_importlib', 'quac/mods.py'),
                           'quac.mods._inspect': ('mods.html#_inspect', 'quac/mods.py'),
                           'quac.mods._itertools': ('mods.html#_itertools', 'quac/mods.py'),
                           'quac.mods._logging': ('mods.html#_logging', 'quac/mods.py'),
                           'quac.mods._math': ('mods.html#_math', 'quac/mods.py'),
                           'quac.mods._matplotlib': ('mods.html#_matplotlib', 'quac/mods.py'),
                           'quac.mods._mpl_artist': ('mods.html#_mpl_artist', 'quac/mods.py'),
                           'quac.mods._mpl_axes': ('mods.html#_mpl_axes', 'quac/mods.py'),
                           'quac.mods._mpl_base': ('mods.html#_mpl_base', 'quac/mods.py'),
                           'quac.mods._mpl_patches': ('mods.html#_mpl_patches', 'quac/mods.py'),
                           'quac.mods._numbers': ('mods.html#_numbers', 'quac/mods.py'),
                           'quac.mods._numpy': ('mods.html#_numpy', 'quac/mods.py'),
                           'quac.mods._os': ('mods.html#_os', 'quac/mods.py'),
                           'quac.mods._pandas': ('mods.html#_pandas', 'quac/mods.py'),
                           'quac.mods._pathlib': ('mods.html#_pathlib', 'quac/mods.py'),
                           'quac.mods._phate': ('mods.html#_phate', 'quac/mods.py'),
                           'quac.mods._pytorch_lightning': ('mods.html#_pytorch_lightning', 'quac/mods.py'),
                           'quac.mods._re': ('mods.html#_re', 'quac/mods.py'),
                           'quac.mods._rich': ('mods.html#_rich', 'quac/mods.py'),
                           'quac.mods._scipy': ('mods.html#_scipy', 'quac/mods.py'),
                           'quac.mods._seaborn': ('mods.html#_seaborn', 'quac/mods.py'),
                           'quac.mods._sklearn': ('mods.html#_sklearn', 'quac/mods.py'),
                           'quac.mods._string': ('mods.html#_string', 'quac/mods.py'),
                           'quac.mods._torch': ('mods.html#_torch', 'quac/mods.py'),
                           'quac.mods._torch_c': ('mods.html#_torch_c', 'quac/mods.py'),
                           'quac.mods._torchdyn': ('mods.html#_torchdyn', 'quac/mods.py'),
                           'quac.mods._torchsde': ('mods.html#_torchsde', 'quac/mods.py'),
                           'quac.mods._torchvision': ('mods.html#_torchvision', 'quac/mods.py'),
                           'quac.mods._tqdm': ('mods.html#_tqdm', 'quac/mods.py'),
                           'quac.mods._typer': ('mods.html#_typer', 'quac/mods.py'),
                           'quac.mods._types': ('mods.html#_types', 'quac/mods.py')},
            'quac.prot': { 'quac.prot.Duckish': ('prot.html#duckish', 'quac/prot.py'),
                           'quac.prot.Duckish.__instancecheck__': ('prot.html#duckish.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.Hashable': ('prot.html#hashable', 'quac/prot.py'),
                           'quac.prot.Hashable.__hash__': ('prot.html#hashable.__hash__', 'quac/prot.py'),
                           'quac.prot.Hashable.__instancecheck__': ('prot.html#hashable.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.NonStrBytesIterable': ('prot.html#nonstrbytesiterable', 'quac/prot.py'),
                           'quac.prot.NonStrBytesIterable.__instancecheck__': ( 'prot.html#nonstrbytesiterable.__instancecheck__',
                                                                                'quac/prot.py'),
                           'quac.prot.NonStrBytesIterable.__iter__': ('prot.html#nonstrbytesiterable.__iter__', 'quac/prot.py'),
                           'quac.prot.NotProtocol': ('prot.html#notprotocol', 'quac/prot.py'),
                           'quac.prot.NotProtocol.__instancecheck__': ('prot.html#notprotocol.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.NotSpecial': ('prot.html#notspecial', 'quac/prot.py'),
                           'quac.prot.NotSpecial.__hash__': ('prot.html#notspecial.__hash__', 'quac/prot.py'),
                           'quac.prot.NotSpecial.__instancecheck__': ('prot.html#notspecial.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.OptProtocol': ('prot.html#optprotocol', 'quac/prot.py'),
                           'quac.prot.OptProtocol.__instancecheck__': ('prot.html#optprotocol.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.Origin': ('prot.html#origin', 'quac/prot.py'),
                           'quac.prot.Origin.__hash__': ('prot.html#origin.__hash__', 'quac/prot.py'),
                           'quac.prot.Origin.__instancecheck__': ('prot.html#origin.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.SpecType': ('prot.html#spectype', 'quac/prot.py'),
                           'quac.prot.SpecType.__instancecheck__': ('prot.html#spectype.__instancecheck__', 'quac/prot.py'),
                           'quac.prot.Special': ('prot.html#special', 'quac/prot.py'),
                           'quac.prot.Special.__hash__': ('prot.html#special.__hash__', 'quac/prot.py'),
                           'quac.prot.Special.__instancecheck__': ('prot.html#special.__instancecheck__', 'quac/prot.py')},
            'quac.solo': { 'quac.solo.Empty': ('solo.html#empty', 'quac/solo.py'),
                           'quac.solo.Module': ('solo.html#module', 'quac/solo.py'),
                           'quac.solo.Singleton': ('solo.html#singleton', 'quac/solo.py'),
                           'quac.solo.Singleton.__call__': ('solo.html#singleton.__call__', 'quac/solo.py'),
                           'quac.solo.Singleton.__eq__': ('solo.html#singleton.__eq__', 'quac/solo.py'),
                           'quac.solo.Singleton.__hash__': ('solo.html#singleton.__hash__', 'quac/solo.py'),
                           'quac.solo.Singleton.__instancecheck__': ('solo.html#singleton.__instancecheck__', 'quac/solo.py'),
                           'quac.solo.Singleton.__new__': ('solo.html#singleton.__new__', 'quac/solo.py'),
                           'quac.solo.Singleton.__repr__': ('solo.html#singleton.__repr__', 'quac/solo.py'),
                           'quac.solo.Singleton.__subclasscheck__': ('solo.html#singleton.__subclasscheck__', 'quac/solo.py')},
            'quac.spec': {},
            'quac.type': {},
            'quac.util': { 'quac.util._fromcall': ('util.html#_fromcall', 'quac/util.py'),
                           'quac.util._fromducks': ('util.html#_fromducks', 'quac/util.py'),
                           'quac.util.asset_attrs': ('util.html#asset_attrs', 'quac/util.py'),
                           'quac.util.asset_last': ('util.html#asset_last', 'quac/util.py'),
                           'quac.util.asset_name': ('util.html#asset_name', 'quac/util.py'),
                           'quac.util.asset_sub_pkg': ('util.html#asset_sub_pkg', 'quac/util.py'),
                           'quac.util.find_asset_module': ('util.html#find_asset_module', 'quac/util.py'),
                           'quac.util.find_last_spec': ('util.html#find_last_spec', 'quac/util.py'),
                           'quac.util.find_spec_or_none': ('util.html#find_spec_or_none', 'quac/util.py'),
                           'quac.util.insys': ('util.html#insys', 'quac/util.py'),
                           'quac.util.load_asset': ('util.html#load_asset', 'quac/util.py'),
                           'quac.util.parse_asset': ('util.html#parse_asset', 'quac/util.py'),
                           'quac.util.prep_kwargs': ('util.html#prep_kwargs', 'quac/util.py'),
                           'quac.util.try_ducked_attr': ('util.html#try_ducked_attr', 'quac/util.py'),
                           'quac.util.try_loaded_attr': ('util.html#try_loaded_attr', 'quac/util.py'),
                           'quac.util.tryattr': ('util.html#tryattr', 'quac/util.py'),
                           'quac.util.unduck': ('util.html#unduck', 'quac/util.py')},
            'quac.zero': { 'quac.zero._acczeros': ('zero.html#_acczeros', 'quac/zero.py'),
                           'quac.zero._format_list': ('zero.html#_format_list', 'quac/zero.py'),
                           'quac.zero._genzeros': ('zero.html#_genzeros', 'quac/zero.py'),
                           'quac.zero._shape': ('zero.html#_shape', 'quac/zero.py'),
                           'quac.zero.empty': ('zero.html#empty', 'quac/zero.py'),
                           'quac.zero.empty.__init__': ('zero.html#empty.__init__', 'quac/zero.py'),
                           'quac.zero.empty.__repr__': ('zero.html#empty.__repr__', 'quac/zero.py'),
                           'quac.zero.empty.ndim': ('zero.html#empty.ndim', 'quac/zero.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_meta.ipynb.

# %% auto 0
__all__ = ['DuckModuleMeta', 'ABCSpecies', 'ABCDuck']

# %% ../nbs/08_meta.ipynb 6
import sys
from abc import ABCMeta, abstractmethod
from functools import wraps 
from inspect import isclass
from importlib import import_module
from importlib.util import (module_from_spec)
from importlib.machinery import ModuleSpec

# %% ../nbs/08_meta.ipynb 8
from types import ModuleType, NoneType
from typing import (Any, Callable, TypeGuard, ClassVar,)


# %% ../nbs/08_meta.ipynb 10
#| export

# %% ../nbs/08_meta.ipynb 12
#| export

# %% ../nbs/08_meta.ipynb 14
from .cons import ANNOYING_ATTRS, MODULE_TYPE_ATTRS
from .type import T, O
from .prot import NotProtocol, OptProtocol
from .enum import Species, LoaderState
from .errs import DuckException, DuckNameError, DuckImportError, DuckAttributeError, DuckModuleNotFoundError
from .solo import Module
from .util import find_last_spec
from .grds import isclstype, isalias, hasduck, hasload

# %% ../nbs/08_meta.ipynb 16
class DuckModuleMeta(ABCMeta, type):
    __duckmodule__: ClassVar[str]
    '''The name of the module to duck.''';
    
    __hush__: ClassVar[bool] = True
    '''Whether or not to raise AttributeError when an attribute is not found.''';
    
    __alts__: ClassVar[dict[str, type]] = {}
    '''Alternatives to use when an attribute is not found.''';
    
    def __instancecheck__(cls: type[T], obj: Any) -> bool:
        '''Check if the object is an instance of the class.'''
        issubbed = isinstance(obj, (ModuleType, DuckModuleMeta))
        samename = object.__getattribute__(cls, '__duckmodule__') == obj.__name__
        hasattrs = all(hasattr(obj, attr) for attr in MODULE_TYPE_ATTRS)
        hasprivs = all(hasattr(obj, attr) for attr in ('___duckmodule__', '__hush__', '__alts__'))
        if issubbed and samename and (hasattrs or hasprivs): return True
        return False
     
    
    def __getattr__(cls: type[T], name: str) -> Any:
        hush = object.__getattribute__(cls, '__hush__')
        try:
            if name in ANNOYING_ATTRS: 
                try:
                    # return super().__getattribute__(name)
                    raise AttributeError(f"Module '{cls.__name__}' has no attribute '{name}'.")
                except:
                    if object.__getattribute__(cls, '__hush__'): return Module
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
            
            # base module name
            base = object.__getattribute__(cls, '__duckmodule__')
            # full module name
            full = f'{base}.{name}' if cls.__name__ != 'DuckModule' else name
            
            mkey = full # key to use for sys.modules
            spec: ModuleSpec | None = None # module spec
            stub: str | None = None
            
            # <base.name> not in sys modules
            if full not in sys.modules:            
                # try using importlib to load the module
                try:
                    module = import_module(full)
                    sys.modules[full] = module
                    mkey = full
                except ModuleNotFoundError: ...
                    
                # try using the last found module spec
                try:
                    spec, stub = find_last_spec(full)
                    module: ModuleType = module_from_spec(spec)
                    sys.modules[stub] = module
                    mkey = stub
                    
                except:
                    if object.__getattribute__(cls, '__hush__'): return Module
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
                
            # <base.name> or <stub> in sys.modules
            module = sys.modules[mkey]
            
            # create a DuckModule subclass for the submodule if needed            
            # @wraps(cls, assigned=('__name__', '__qualname__', '__doc__', '__annotations__'), updated=())
            # class ducked(cls):
            #     __duckmodule__ = mkey
                
            ducked = type(mkey, (cls, ), dict(__duckmodule__=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,))
            # ducked = ducked(module.__name__, module.__doc__)
            # ducked.__dict__ = {**module.__dict__, **ducked.__dict__}
            # ducked.__dict__.update({ **module.__dict__, **ducked.__dict__,**dict(_duck=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,)})

            # Try to get the attribute from the module if it exists
            try: 
                attr = object.__getattribute__(module, name)
                if LoaderState.get(module) not in [LoaderState.NONE, LoaderState.DUCKED]:
                    sys.modules[mkey] = LoaderState.set(module, LoaderState.EXECED)
                return attr
            except AttributeError: ...
            
            # Try actually loading the module if it hasn't been loaded yet
            try: 
                sys.modules[mkey] = LoaderState.exec(module)
                return object.__getattribute__(module, name)
            except AttributeError: ...
            
            # Handle fallbacks if defined
            if hasattr(cls, '__alts__') and name in cls.__alts__: return cls.__alts__[name]
            
            return ducked
        
        except NameError as e:
            if not hush: raise DuckNameError(*e.args)
        
        except ModuleNotFoundError as e:
            if not hush: raise DuckModuleNotFoundError(e.args, e.name, e.path)
        
        except ImportError as e:
            if not hush: raise DuckImportError(e.args, e.name, e.path)
        
        except AttributeError as e:
            if not hush: raise DuckAttributeError(e.args, e.name, e.obj)
        
        except Exception as e:
            if not hush: raise DuckException(f'NameError: {e}')
        
        return Module

# %% ../nbs/08_meta.ipynb 18
class ABCSpecies(ABCMeta):
    __species__: ClassVar[Species] = Species.REG
    '''The specific "species" of the protocol. Defaults to `REG` (regular).'''
        
    def __reg__(cls) -> type[T]:
        '''Set the "species" to `REG` (regular).'''
        cls.__species__ = Species.REG
        return cls
    
    def __not__(cls) -> Species:
        '''Set the "species" to `NOT` (negated).'''
        cls.__species__ = Species.NOT
        return cls
    
    def __opt__(cls) -> Species:
        '''Set the "species" to `OPT` (optional).'''
        cls.__species__ = Species.OPT
        return cls
    
    def __migrate__(cls, species: Species) -> type[T]:
        '''Migrate the protocol to a different "species".'''
        match species:
            case Species.REG: return cls.__reg__()
            case Species.NOT: return cls.__not__()
            case Species.OPT: return cls.__opt__()
            case _: return cls # NOTE: fail silently
        return cls

# %% ../nbs/08_meta.ipynb 20
# class DuckMeta(_ProtocolMeta, ABCSpecies, type):
class ABCDuck(ABCSpecies, type):
    __ducktype__: ClassVar[tuple[type[T], ...]] = tuple()
    '''The specific types that the protocol is duck-typed to.''';
    
    __expected__: ClassVar[dict[str, Any]] = dict()
    '''Dictionary of attributes and their expected values to check in the object.''';
    
    __required__: ClassVar[tuple[str, ...]] = tuple()
    '''Iterable of attribute names that the object must have.''';
    
    __allequal__: ClassVar[tuple[str, ...]] = tuple(('type', 'dtype', 'types'))
    '''Iterable of attributes to check if their values are all the same in an iterable object.''';

    __extrafns__: ClassVar[dict[str, Callable[[O, T], TypeGuard[T]]]] = dict()
    '''Mapping of attribute names to guard functions for additional checks.''';
    
    __usealias__: ClassVar[bool] = True
    '''Whether or not to use the `isalias` function for instance checking or `__guard__`.''';
    
    @property
    def __types__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to.'''
        types = []
        for t in cls.__ducktype__:
            if hasduck(t): 
                try:
                    types.append(t.duck)
                    continue
                except Exception: ...
            if hasload(t): 
                try:
                    types.append(t.load())
                except Exception: ...
                
            if isclstype(t): 
                types.append(t)
            else: 
                types.append(type(t))
        
        return tuple(types)
    
    @property
    def __typesq__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to, including NoneType.'''
        return cls.__types__ + (NoneType, )
    
    @property
    def __usetypes__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed and to use.'''
        if cls.__species__ == Species.OPT: return cls.__typesq__
        return cls.__types__
    
    def __guard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol.'''
        return isinstance(inst, cls.__types__)
    
    def __guardq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol, including NoneType.'''
        return isinstance(inst, cls.__typesq__)
    
    def __alias__(cls: type[T], inst: Any) -> TypeGuard[type[T]]:
        return isalias(
            inst, 
            types = cls.__ducktype__, 
            # guards = (cls.__guard__, ),
            attrs = cls.__expected__, 
            hasattrs = cls.__required__, 
            allattrs = cls.__allequal__, 
            mapattrs = cls.__extrafns__
        )
        
    def __instancecheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        if cls.__usealias__: 
            return cls.__alias__(inst)
        match cls.__species__:
            case Species.REG: return cls.__guard__(inst)
            case Species.NOT: return cls.__notinstance__(inst)
            case Species.OPT: return cls.__optinstance__(inst)
            case _: return cls.__guard__(inst)
        return cls.__guard__(inst)
    
    def __subclasscheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return issubclass(inst, cls.__types__)
            case Species.NOT: return not issubclass(inst, cls.__types__)
            case Species.OPT: return issubclass(inst, cls.__typesq__)
        return issubclass(inst, cls.__types__)
    
    def __notinstance__(cls, inst: Any) -> TypeGuard[NotProtocol[T]]:
        '''Return a type guard for NOT the protocol.'''
        return not cls.__instancecheck__(inst)
    
    def __optinstance__(cls, inst: Any) -> TypeGuard[OptProtocol[T]]:
        '''Return a type guard for the protocol or NoneType.'''
        return cls.__guardq__(inst)
    
    def __invert__(cls) -> NotProtocol[T]:
        '''Return the NOT type guard for the protocol.'''
        return cls.__notinstance__
    
    def __memberguard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        types = cls.__types__
        if cls.__species__ == Species.OPT: types = cls.__typesq__
        def checkmember(inst: Any) -> TypeGuard[type[T]]:
            return bool((inst in types) ^ (cls.__species__ == Species.NOT))
        return checkmember
    
    def __membercheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return inst in cls.__types__
            case Species.NOT: return not inst in cls.__types__
            case Species.OPT: return inst in cls.__typesq__
            case _: return inst in cls.__types__
    
    def __contains__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        return cls.__membercheck__(inst)
    
    def __eq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Equality check for the protocol using custom `__guard__` method'''
        try: 
            if cls.__instancecheck__(inst): 
                return True
        except TypeError as inserror: ...
        try: 
            if cls.__subclasscheck__(inst):
                return True
        except TypeError as suberror: ...
        try: 
            if inst in cls.__types__:
                return True
        except TypeError as inerror: ...
        try: 
            if inst is cls:
                return True
        except TypeError as iserror: ...
        return False
    
    
    def __repr__(cls) -> TypeGuard[type[T]]:
        name = cls.__name__
        base = f'<duck {name}>'
        match cls.__species__:
            case Species.REG: return f'<duck {name}>'
            case Species.NOT: return f'<!duck {name}>'
            case Species.OPT: return f'<?duck {name}>'
            
            
    def __call__(cls, *args, **kwargs) -> TypeGuard[type[T]]:
        '''Return the protocol class itself.'''
        species = kwargs.get('species', cls.__species__)
        match species:
            case Species.REG: fn = cls.__guard__
            case Species.NOT: fn = cls.__notinstance__
            case Species.OPT: fn = cls.__optinstance__
            case _: fn = cls.__guard__
        results = [fn(arg) for arg in args]
        if len(results) == 1: return results[0]
        return tuple(results)
            
        
